"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import os
import moderngl
import moderngl_window
import imgui
import shutil
import tempfile
import numpy as np
import struct

from array import array
from aitviewer.configuration import CONFIG as C
from aitviewer.scene.camera import PinholeCamera
from aitviewer.scene.scene import Scene
from aitviewer.streamables.streamable import Streamable
from aitviewer.utils import images_to_video
from collections import namedtuple
from moderngl_window import activate_context
from moderngl_window import geometry
from moderngl_window import get_local_window_cls
from moderngl_window.integrations.imgui import ModernglWindowRenderer
from moderngl_window.timers.clock import Timer
from moderngl_window.opengl.vao import VAO
from pathlib import Path
from PIL import Image
from tqdm import tqdm


MeshMouseIntersection = namedtuple('MeshMouseIntersection', 'node tri_id vert_id point_world point_local bc_coords')


class Viewer(moderngl_window.WindowConfig):
    resource_dir = Path(__file__).parent / 'shaders'
    window_type = 'pyqt5'
    size_mult = 1.0
    samples = 4

    def __init__(self, title="AITViewer", **kwargs):
        """
        Initializer.
        :param title: Window title
        :param kwargs: kwargs.
        """

        # Window Setup (Following `moderngl_window.run_window_config`).
        base_window_cls = get_local_window_cls(self.window_type)

        # Calculate window size
        size = int(C.window_width * self.size_mult), int(C.window_height * self.size_mult)

        self.window = base_window_cls(
            title=title,
            size=size,
            fullscreen=C.fullscreen,
            resizable=C.resizable,
            gl_version=(3, 3),
            aspect_ratio=None,  # Have to set this to None otherwise the window will enforce this aspect ratio.
            vsync=C.vsync,  # Set to False for some performance gains.
            samples=self.samples,
            cursor=True,
        )
        self.window.print_context_info()
        activate_context(window=self.window)
        # self.window.config = self
        self.timer = Timer()
        self.ctx = self.window.ctx
        super().__init__(self.ctx, self.window, self.timer)

        # Create GUI context
        imgui.create_context()
        self.imgui = ModernglWindowRenderer(self.wnd)
        self.imgui_user_interacting = False

        # Setup scene
        self.scene = Scene()
        self.scene.camera = PinholeCamera(45.0)

        # Setup shadow mapping
        offscreen_size = 8192, 8192
        self.offscreen_depth = self.ctx.depth_texture(offscreen_size)
        self.offscreen_depth.compare_func = ''
        self.offscreen_depth.repeat_x = False
        self.offscreen_depth.repeat_y = False
        self.offscreen_color = self.ctx.texture(offscreen_size, 4)
        self.offscreen = self.ctx.framebuffer(
            color_attachments=[self.offscreen_color],
            depth_attachment=self.offscreen_depth
        )
        # Shaders for rendering the shadow map
        self.raw_depth_prog = self.load_program('shadow_mapping/raw_depth.glsl')
        self.shadowmap_prog = self.load_program('shadow_mapping/shadowmap.glsl')

        # Mesh mouse intersection
        self.offscreen_p_depth = self.ctx.depth_texture(self.wnd.buffer_size)
        self.offscreen_p_viewpos = self.ctx.texture(self.wnd.buffer_size, 4, dtype='f4')
        self.offscreen_p_tri_id = self.ctx.texture(self.wnd.buffer_size, 4, dtype='f4')
        self.offscreen_p = self.ctx.framebuffer(
            color_attachments=[
                self.offscreen_p_viewpos,
                self.offscreen_p_tri_id
            ],
            depth_attachment=self.offscreen_p_depth
        )
        # Shaders for mesh mouse intersection
        self.frag_map_prog = self.load_program('fragment_picking/frag_map.glsl')
        self.frag_pick_prog = self.load_program('fragment_picking/frag_pick.glsl')
        self.frag_pick_prog['position_texture'].value = 0  # Read from texture channel 0
        self.frag_pick_prog['obj_info_texture'].value = 1  # Read from texture channel 0
        self.picker_output = self.ctx.buffer(reserve=5*4)  # 3 floats, 2 ints
        self.picker_vao = VAO(mode=moderngl.POINTS)

        # For debugging
        self.offscreen_quad = geometry.quad_2d(size=(0.5, 0.5), pos=(0.75, 0.75))

        # Custom UI Font
        self.font_dir = Path(__file__).parent / 'resources' / 'fonts'
        self.fonts = imgui.get_io().fonts
        self.custom_font = self.fonts.add_font_from_file_ttf(os.path.join(self.font_dir, 'Custom.ttf'), 15)
        self.scene.custom_font = self.custom_font
        self.imgui.refresh_font_texture()

        self.modes = {
            'view': {'title': ' View', 'shortcut': 'V'},
            'inspect': {'title': ' Inspect', 'shortcut': 'I'},
        }
        self.selected_mode = 'view'

        self.gui_controls = {
            'menu': self.gui_menu,
            'scene': self.gui_scene,
            'playback': self.gui_playback,
            'inspect': self.gui_inspect
        }

        # Settings
        self.run_animations = C.run_animations
        self.draw_edges = C.draw_edges
        self.flat_rendering = C.flat_rendering
        self.dark_mode = C.dark_mode
        self.playback_fps = C.playback_fps
        self.shadows_enabled = C.shadows_enabled
        self.show_shadow_map = C.show_shadow_map
        self.auto_set_floor = C.auto_set_floor
        self.auto_set_camera_target = C.auto_set_camera_target
        self.backface_culling = C.backface_culling

        self._pan_camera = False
        self._rotate_camera = False
        self._past_framerates = np.zeros([60]) - 1.0
        self._last_frame_rendered_at = 0

        self.mmi = None  # Mesh mouse intersection result
        self.animation_range = [0, -1]
        self.video_as_gif = False
        self.video_rotate = False

        # Key Shortcuts
        self._pause_key = self.wnd.keys.SPACE
        self._next_frame_key = self.wnd.keys.PERIOD
        self._previous_frame_key = self.wnd.keys.COMMA
        self._draw_edges_key = self.wnd.keys.E
        self._flat_rendering_key = self.wnd.keys.F
        self._shadow_key = self.wnd.keys.S
        self._orthographic_camera_key = self.wnd.keys.O
        self._mode_inspect = self.wnd.keys.I
        self._mode_view = self.wnd.keys.V
        self._dark_mode_key = self.wnd.keys.D
        self._screenshot_key = self.wnd.keys.P
        self._right_mouse_button = 2  # right
        self._left_mouse_button = 1  # left
        self._save_cam_key = self.wnd.keys.C
        self._load_cam_key = self.wnd.keys.L
        self._shortcut_names = {self.wnd.keys.SPACE: "Space",
                                self.wnd.keys.C: "C",
                                self.wnd.keys.D: "D",
                                self.wnd.keys.E: "E",
                                self.wnd.keys.I: "I",
                                self.wnd.keys.F: "F",
                                self.wnd.keys.L: "L",
                                self.wnd.keys.O: "O",
                                self.wnd.keys.P: "P",
                                self.wnd.keys.S: "S"}

    def run(self, *args, log=True):
        """
        Enter a blocking visualization loop. This is built following `moderngl_window.run_window_config`.
        :param args: The arguments passed to `config_cls` constructor.
        :param log: Whether to log to the console.
        """
        self.scene.make_renderable(self.ctx)
        if self.auto_set_floor:
            self.scene.auto_set_floor()
            self.scene.backface_culling = self.backface_culling
        if self.auto_set_camera_target:
            self.scene.auto_set_camera_target()
        self.scene.set_lights(self.dark_mode)

        self.animation_range[-1] = self.scene.n_frames - 1

        self.timer.start()
        while not self.window.is_closing:
            current_time, delta = self.timer.next_frame()
            self.window.clear()
            self.window.render(current_time, delta)
            self.window.swap_buffers()
        _, duration = self.timer.stop()
        self.on_close()
        self.window.destroy()
        if duration > 0 and log:
            print("Duration: {0:.2f}s @ {1:.2f} FPS".format(duration, self.window.frames / duration))

    def render(self, time, frame_time):
        """The main drawing function."""
        # Check if we need to advance the sequences.
        if self.run_animations and time - self._last_frame_rendered_at > 1.0 / self.playback_fps:
            self.scene.next_frame()
            self._last_frame_rendered_at = time

        self.streamable_capture()
        self.render_fragmap()
        self.render_shadowmap()
        self.render_prepare()
        self.render_scene()

        if self.show_shadow_map:
            self.offscreen_p_depth.use(location=0)
            self.offscreen_quad.render(self.frag_map_prog)

        # FPS accounting.
        self._past_framerates[:-1] = self._past_framerates[1:]
        self._past_framerates[-1] = 1 / frame_time if frame_time > 0 else 9000.0

        # Render the UI components.
        self.gui()

    def streamable_capture(self):
        # Collect all streamable nodes
        rs = self.scene.collect_nodes(obj_type=Streamable)
        for r in rs:
            r.capture()

    def render_shadowmap(self):
        """A pass to render the shadow map, i.e. render the entire scene once from the view of the light."""
        self.ctx.enable_only(moderngl.DEPTH_TEST)
        self.offscreen.clear()
        self.offscreen.use()
        if self.shadows_enabled:
            rs = self.scene.collect_nodes()
            for r in rs:
                r.render_shadowmap(self.scene.lights[0].mvp(), self.shadowmap_prog)

    def render_fragmap(self):
        """A pass to render the fragment picking map, i.e. render the scene with world coords as colors."""
        self.ctx.enable_only(moderngl.DEPTH_TEST)
        self.offscreen_p.clear()
        self.offscreen_p.use()
        rs = self.scene.collect_nodes()
        for r in rs:
            r.render_fragmap(self.scene.camera, self.frag_map_prog, self.wnd.buffer_size)

    def render_scene(self):
        """Render the current scene to the framebuffer without time accounting and GUI elements."""
        self.scene.render(window_size=self.window.size,
                          draw_edges=self.draw_edges,
                          flat_rendering=self.flat_rendering,
                          offscreen_depth=self.offscreen_depth,
                          light_mvp=self.scene.lights[0].mvp(),
                          shadows_enabled=self.shadows_enabled)

    def render_prepare(self):
        """Prepare the framebuffer."""
        self.wnd.use()
        # Clear background and make sure only the flags we want are enabled.
        if self.dark_mode:
            self.ctx.clear(0.1, 0.1, 0.1, 1.0)
        else:
            self.ctx.clear(1.0, 1.0, 1.0, 1.0)

        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.BLEND | moderngl.CULL_FACE)
        self.ctx.cull_face = 'back'
        self.ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA

    def prevent_background_interactions(self):
        """Prevent background interactions when hovering over any imgui window."""
        imgui_hover = any([imgui.is_window_hovered(),
                           imgui.is_any_item_hovered(),
                           ])
        self.imgui_user_interacting = True if imgui_hover else self.imgui_user_interacting

    def gui(self):
        imgui.new_frame()

        # Reset user interacting state
        self.imgui_user_interacting = False

        # Render user controls
        for gc in self.gui_controls.values(): gc()

        # Contains live examples of all possible displays/controls - useful for browsing for new components
        # imgui.show_test_window()

        imgui.render()
        self.imgui.render(imgui.get_draw_data())

    def gui_scene(self):
        # Render scene GUI
        imgui.begin(self.scene.name, True)
        self.scene.gui(imgui)
        self.prevent_background_interactions()
        imgui.end()

    def gui_menu(self):
        clicked_export = False

        if imgui.begin_main_menu_bar():
            if imgui.begin_menu("File", True):

                clicked_quit, selected_quit = imgui.menu_item(
                    "Quit", 'Cmd+Q', False, True)
                if clicked_quit:
                    exit(1)

                clicked_export, selected_export = imgui.menu_item(
                    "Save as video..", None, False, True)

                clicked_export360, _ = imgui.menu_item(
                    "Save as 360 video..", None, False, True)
                if clicked_export360:
                    self.to_360_deg_video()

                clicked_screenshot, selected_screenshot = imgui.menu_item(
                    "Screenshot", self._shortcut_names[self._screenshot_key], False, True)
                if clicked_screenshot:
                    self.take_screenshot()

                imgui.end_menu()

            if imgui.begin_menu("View", True):
                _, self.flat_rendering = imgui.menu_item("Flat shading", self._shortcut_names[self._flat_rendering_key],
                                                         self.flat_rendering, True)
                _, self.draw_edges = imgui.menu_item("Draw edges", self._shortcut_names[self._draw_edges_key],
                                                     self.draw_edges, True)
                _, self.shadows_enabled = imgui.menu_item("Render Shadows", self._shortcut_names[self._shadow_key],
                                                          self.shadows_enabled, True)
                # _, self.show_shadow_map = imgui.menu_item("Show Shadow Map", None, self.show_shadow_map, True)
                _, self.dark_mode = imgui.menu_item("Dark Mode", self._shortcut_names[self._dark_mode_key],
                                                    self.dark_mode, True)
                _, self.scene.camera.is_ortho = imgui.menu_item("Orthographic Camera",
                                                                self._shortcut_names[self._orthographic_camera_key],
                                                                self.scene.camera.is_ortho, True)
                clicked_save_cam, selected_save_cam = imgui.menu_item("Save Camera",
                                                                self._shortcut_names[self._save_cam_key],
                                                                False, True)
                if clicked_save_cam:
                    self.scene.camera.save_cam()

                clicked_load_cam, selected_load_cam = imgui.menu_item("Load Camera",
                                                                self._shortcut_names[self._load_cam_key],
                                                                False, True)

                if clicked_load_cam:
                    self.scene.camera.load_cam()

                imgui.end_menu()

            if imgui.begin_menu("Mode", True):
                for id, mode in self.modes.items():
                    mode_clicked, _ = imgui.menu_item(mode['title'], mode['shortcut'], id == self.selected_mode, True)
                    if mode_clicked:
                        self.selected_mode = id

                imgui.end_menu()
            imgui.end_main_menu_bar()

        if clicked_export:
            imgui.open_popup("Export Video")
        if imgui.begin_popup_modal("Export Video")[0]:
            imgui.text("Enter Range for Video Export.")
            rcu, animation_range = imgui.drag_int2('Range##range_control', *self.animation_range,
                                                   min_value=0,  max_value=self.scene.n_frames-1)
            if rcu:
                self.animation_range[0] = max(animation_range[0], 0)
                self.animation_range[1] = min(animation_range[-1], self.scene.n_frames-1)
            _, self.video_as_gif = imgui.checkbox("GIF instead of MP4", self.video_as_gif)
            _, self.video_rotate = imgui.checkbox("Animate and rotate", self.video_rotate)
            if imgui.button("OK"):
                imgui.close_current_popup()
                # Export with applied settings
                self.to_video(self.video_rotate)
            self.prevent_background_interactions()
            imgui.end_popup()

    def gui_playback(self):
        """GUI to control playback settings."""
        imgui.begin("Playback", True)
        _, self.run_animations = imgui.checkbox("Run animations [{}]".format(self._shortcut_names[self._pause_key]),
                                                self.run_animations)

        # Plot FPS
        fps_avg = np.mean(self._past_framerates[self._past_framerates > 0.0])
        ms_avg = 1000.0 / fps_avg

        imgui.plot_lines("Internal {:.1f} fps @ {:.2f} ms/frame".format(fps_avg, ms_avg),
                         array('f', self._past_framerates.tolist()),
                         scale_min=0, scale_max=100.0, graph_size=(100, 20))

        _, self.playback_fps = imgui.drag_float('Target Playback fps##playback_fps', self.playback_fps, 0.1,
                                                min_value=0.0, max_value=120.0, format='%.1f')

        # Sequence Control
        # For simplicity, we allow the global sequence slider to only go as far as the shortest known sequence.
        n_frames = self.scene.n_frames

        _, self.scene.current_frame_id = imgui.slider_int('Frame##r_global_seq_control', self.scene.current_frame_id,
                                                          min_value=0, max_value=n_frames - 1)
        self.prevent_background_interactions()
        imgui.end()

    def gui_inspect(self):
        """GUI to control playback settings."""
        if self.selected_mode == 'inspect':
            imgui.begin("Inspect", True)

            if self.mmi is not None:
                for k, v in zip(self.mmi._fields, self.mmi):
                    imgui.text("{}: {}".format(k, v))

            self.prevent_background_interactions()
            imgui.end()

    def mesh_mouse_intersection(self, x: int, y: int):
        """Given an x/y screen coordinate, get the intersected object, triangle id, and xyz point in camera space"""

        # Texture is y=0 at bottom, so we flip y coords
        pos = int(x * self.wnd.pixel_ratio), int(self.wnd.buffer_height - (y * self.wnd.pixel_ratio))

        # Fragment picker uses already encoded position/object/triangle in the frag_pos program textures
        self.frag_pick_prog['texel_pos'].value = pos
        self.offscreen_p_viewpos.use(location=0)
        self.offscreen_p_tri_id.use(location=1)
        self.picker_vao.transform(self.frag_pick_prog, self.picker_output, vertices=1)
        x, y, z, obj_id, tri_id = struct.unpack('3f2i', self.picker_output.read())

        if obj_id > 0 and tri_id > 0:
            node = self.scene.get_node_by_uid(obj_id)
            # Camera space to world space
            point_world = np.array(np.linalg.inv(self.scene.camera.view()) @ np.array((x, y, z, 1.0)))[:-1]
            point_local = (np.linalg.inv(node.model_matrix()) @ np.append(point_world, 1.0))[:-1]
            vert_id = node.closest_vertex_in_triangle(tri_id, point_local)
            bc_coords = node.get_bc_coords_from_points(tri_id, [point_local])
            return MeshMouseIntersection(node, tri_id, vert_id, point_world, point_local, bc_coords)

        return None

    def resize(self, width: int, height: int):
        self.window_size = (width, height)
        self.imgui.resize(width, height)

    def key_event(self, key, action, modifiers):
        self.imgui.key_event(key, action, modifiers)
        if action == self.wnd.keys.ACTION_PRESS:

            if key == self._pause_key:
                self.run_animations = not self.run_animations

            elif key == self._next_frame_key:
                if not self.run_animations:
                    self.scene.next_frame()

            elif key == self._previous_frame_key:
                if not self.run_animations:
                    self.scene.previous_frame()

            elif key == self._draw_edges_key:
                self.draw_edges = not self.draw_edges

            elif key == self._shadow_key:
                self.shadows_enabled = not self.shadows_enabled

            elif key == self._orthographic_camera_key:
                self.scene.camera.is_ortho = not self.scene.camera.is_ortho

            elif key == self._flat_rendering_key:
                self.flat_rendering = not self.flat_rendering

            elif key == self._mode_view:
                self.selected_mode = 'view'

            elif key == self._mode_inspect:
                self.selected_mode = 'inspect'

            elif key == self._dark_mode_key:
                self.dark_mode = not self.dark_mode
                self.scene.set_lights(self.dark_mode)

            elif key == self._screenshot_key:
                self.take_screenshot()
            elif key == self._save_cam_key:
                self.scene.camera.save_cam()
            elif key == self._load_cam_key:
                self.scene.camera.load_cam()

        if action == self.wnd.keys.ACTION_RELEASE:
            pass

    def mouse_position_event(self, x, y, dx, dy):
        self.imgui.mouse_position_event(x, y, dx, dy)

        if self.selected_mode == 'inspect':
            self.mmi = self.mesh_mouse_intersection(x, y)

    def mouse_press_event(self, x: int, y: int, button: int):
        self.imgui.mouse_press_event(x, y, button)

        if not self.imgui_user_interacting:
            self._pan_camera = button == self._right_mouse_button
            self._rotate_camera = button == self._left_mouse_button

    def mouse_release_event(self, x: int, y: int, button: int):
        self.imgui.mouse_release_event(x, y, button)

        if button == self._right_mouse_button:
            self._pan_camera = False
        if button == self._left_mouse_button:
            self._rotate_camera = False

    def mouse_drag_event(self, x: int, y: int, dx: int, dy: int):
        self.imgui.mouse_drag_event(x, y, dx, dy)

        if not self.imgui_user_interacting :
            if self._pan_camera:
                self.scene.camera.pan(dx, dy)

            if self._rotate_camera:
                self.scene.camera.rotate_azimuth_elevation(dx, dy)

    def mouse_scroll_event(self, x_offset: float, y_offset: float):
        self.imgui.mouse_scroll_event(x_offset, y_offset)

        if not self.imgui_user_interacting:
            self.scene.camera.dolly_zoom(np.sign(y_offset))

    def unicode_char_entered(self, char):
        self.imgui.unicode_char_entered(char)

    def save_current_frame_as_image(self, frame_dir, frame_id):
        """Saves the current frame as an image to disk."""
        image = self.get_current_frame_as_image()
        image_name = os.path.join(frame_dir, 'frame_{:0>6}.png'.format(frame_id))
        image.save(image_name)

    def get_current_frame_as_image(self):
        """Return the FBO content as a PIL image."""
        image = Image.frombytes('RGB',
                                (self.wnd.fbo.viewport[2] - self.wnd.fbo.viewport[0],
                                 self.wnd.fbo.viewport[3] - self.wnd.fbo.viewport[1]),
                                self.wnd.fbo.read(viewport=self.wnd.fbo.viewport, alignment=1))
        image = image.transpose(Image.FLIP_TOP_BOTTOM)
        return image

    def on_close(self):
        """
        Clean up before destroying the window
        """
        # Shut down all streams
        for s in self.scene.collect_nodes(obj_type=Streamable):
            s.stop()

    def take_screenshot(self):
        """Save the current framebuffer to an image."""
        frame_dir = C.export_dir + '/screenshots/'
        if not os.path.exists(frame_dir):
            os.makedirs(frame_dir)
        # We don't want to export GUI elements, so render the scene again.
        self.render_shadowmap()
        self.render_prepare()
        self.render_scene()
        self.save_current_frame_as_image(frame_dir, self.scene.current_frame_id)

    def to_video(self, rotate=False):
        """Render the loaded animations to video."""

        if self.animation_range[-1] < self.animation_range[0]:
            print("No frames rendered.")
            return

        # Start all sequences from 0 but remember where we left off.
        saved_curr_frame = self.scene.current_frame_id

        # Create a temporary directory to render frames to.
        frame_dir = tempfile.TemporaryDirectory().name
        os.makedirs(frame_dir)

        saved_camera = copy.deepcopy(self.scene.camera)
        az_delta = 2 * np.pi / (self.animation_range[-1]+1) if rotate else 0.0

        # Render each frame to an image file.
        print("Saving frames to {}".format(frame_dir))
        progress_bar = tqdm(total=self.animation_range[-1]-self.animation_range[0], desc='Rendering frames')

        for i, f in enumerate(range(self.animation_range[0], self.animation_range[-1]+1)):
            self.scene.camera.rotate_azimuth(az_delta)
            self.scene.current_frame_id = f
            self.render_shadowmap()
            self.render_prepare()
            self.render_scene()
            self.save_current_frame_as_image(frame_dir, i)
            progress_bar.update()
        progress_bar.close()

        # Export to video.
        suffix = '.gif' if self.video_as_gif else '.mp4'
        images_to_video(frame_dir, C.export_dir + '/videos/' + self.window.title + suffix, input_fps=self.playback_fps,
                        output_fps=60)

        # Remove temp frames.
        shutil.rmtree(frame_dir)

        # Reset frame IDs.
        self.scene.current_frame_id = saved_curr_frame
        self.scene.camera = saved_camera

        print("Done.")

    def to_360_deg_video(self):
        """Matrix shot. Keeps the look-at point fixed and rotates the camera around it."""
        n_frames = 360
        saved_camera = copy.deepcopy(self.scene.camera)

        angles = -np.arange(0, 2*np.pi, 2*np.pi/n_frames)
        cam_centered = self.scene.camera.position - self.scene.camera.target
        new_x = cam_centered[0] * np.cos(angles) - cam_centered[2] * np.sin(angles) + self.scene.camera.target[0]
        new_z = cam_centered[0] * np.sin(angles) + cam_centered[2] * np.cos(angles) + self.scene.camera.target[2]
        circle = np.stack([new_x, np.zeros_like(new_x) + self.scene.camera.position[1], new_z]).T

        # Create a temporary directory to render frames to.
        frame_dir = tempfile.TemporaryDirectory().name
        os.makedirs(frame_dir)

        # Render each frame to an image file.
        print("Saving frames to {}".format(frame_dir))
        progress_bar = tqdm(total=n_frames, desc='Rendering frames')
        for f in range(circle.shape[0]):
            self.scene.camera.position = circle[f]
            self.render_shadowmap()
            self.render_prepare()
            self.render_scene()
            self.save_current_frame_as_image(frame_dir, f)
            progress_bar.update()
        progress_bar.close()

        # Export to video.
        images_to_video(frame_dir, C.export_dir + '/videos/' + self.window.title + '_360deg.mp4')

        # Remove temp frames.
        shutil.rmtree(frame_dir)

        # Reset Camera.
        self.scene.camera = saved_camera

        print("Done.")



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import pickle as pkl
import torch

from aitviewer.configuration import CONFIG as C
from aitviewer.models.smpl import SMPLLayer
from aitviewer.renderables.meshes import Meshes
from aitviewer.renderables.skeletons import Skeletons
from aitviewer.renderables.rigid_bodies import RigidBodies
from aitviewer.scene.node import Node
from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa
from aitviewer.utils.so3 import interpolate_rotations
from aitviewer.utils.so3 import resample_rotations
from aitviewer.utils import resample_positions
from aitviewer.utils import to_torch
from aitviewer.utils import local_to_global
from aitviewer.utils import interpolate_positions
from aitviewer.utils import to_numpy as c2c
from scipy.spatial.transform import Rotation as R


class SMPLSequence(Node):
    """
    Represents a temporal sequence of SMPL poses. Can be loaded from disk or initialized from memory.
    """

    def __init__(self,
                 poses_body,
                 smpl_layer,
                 poses_root=None,
                 betas=None,
                 trans=None,
                 poses_left_hand=None,
                 poses_right_hand=None,
                 device=C.device,
                 dtype=C.f_precision,
                 include_root=True,
                 normalize_root=False,
                 is_rigged=True,
                 show_joint_angles=False,
                 **kwargs):
        """
        Initializer.
        :param poses_body: An array (numpy ar pytorch) of shape (F, N_JOINTS*3) containing the pose parameters of the
          body, i.e. without hands or face parameters.
        :param smpl_layer: The SMPL layer that maps parameters to joint positions and/or dense surfaces.
        :param poses_root: An array (numpy or pytorch) of shape (F, 3) containing the global root orientation.
        :param betas: An array (numpy or pytorch) of shape (N_BETAS, ) containing the shape parameters.
        :param trans: An array (numpy or pytorch) of shape (F, 3) containing the global root translation.
        :param device: The pytorch device for computations.
        :param dtype: The pytorch data type.
        :param include_root: Whether or not to include root information. If False, no root translation and no root
          rotation is applied.
        :param normalize_root: Whether or not to normalize the root. If True, the global root translation in the first
          frame is zero and the global root orientation is the identity.
        :param is_rigged: Whether or not to display the joints as a skeleton.
        :param show_joint_angles: Whether or not the coordinate frames at the joints should be visualized.
        :param kwargs: Remaining arguments for rendering.
        """
        assert len(poses_body.shape) == 2
        super(SMPLSequence, self).__init__(n_frames=poses_body.shape[0], **kwargs)

        self.smpl_layer = smpl_layer

        self.poses_body = to_torch(poses_body, dtype=dtype, device=device)
        self.poses_left_hand = to_torch(poses_left_hand, dtype=dtype, device=device)
        self.poses_right_hand = to_torch(poses_right_hand, dtype=dtype, device=device)

        poses_root = poses_root if poses_root is not None else torch.zeros([len(poses_body), 3])
        betas = betas if betas is not None else torch.zeros([1, self.smpl_layer.num_betas])
        trans = trans if trans is not None else torch.zeros([len(poses_body), 3])

        self.poses_root = to_torch(poses_root, dtype=dtype, device=device)
        self.betas = to_torch(betas, dtype=dtype, device=device)
        self.trans = to_torch(trans, dtype=dtype, device=device)

        if len(self.betas.shape) == 1:
            self.betas = self.betas.unsqueeze(0)

        self._include_root = include_root
        self._normalize_root = normalize_root
        self._show_joint_angles = show_joint_angles
        self._is_rigged = is_rigged or show_joint_angles
        self._render_kwargs = kwargs

        if not self._include_root:
            self.poses_root = torch.zeros_like(self.poses_root)
            self.trans = torch.zeros_like(self.trans)

        if self._normalize_root:
            root_ori = aa2rot(self.poses_root)
            first_root_ori = torch.inverse(root_ori[0:1])
            root_ori = torch.matmul(first_root_ori, root_ori)
            self.poses_root = rot2aa(root_ori)

            trans = torch.matmul(first_root_ori.unsqueeze(0), self.trans.unsqueeze(-1)).squeeze()
            self.trans = trans - trans[0:1]

        # Nodes
        self.vertices, self.joints, self.faces, self.skeleton = self.fk()

        if self._is_rigged:
            # Must first add skeleton, otherwise transparency does not work correctly.
            # Overriding given color with a custom color for the skeleton.
            kwargs = self._render_kwargs.copy()
            kwargs['color'] = (1.0, 177 / 255, 1 / 255, 1.0)
            kwargs['name'] = 'Skeleton'
            self.skeleton_seq = Skeletons(self.joints, self.skeleton, **kwargs)
            self.skeleton_seq.position = self.position
            self.skeleton_seq.rotation = self.rotation

            self._add_node(self.skeleton_seq, gui_elements=['material'])

        if self._show_joint_angles:
            # First convert the relative joint angles to global joint angles in rotation matrix form.
            global_oris = local_to_global(torch.cat([self.poses_root, self.poses_body], dim=-1),
                                          self.skeleton[:, 0], output_format='rotmat')
            global_oris = global_oris.reshape((self.n_frames, -1, 3, 3))

            self.rbs = RigidBodies(self.joints, c2c(global_oris), length=0.1, name='Joint Angles')
            self.rbs.position = self.position
            self.rbs.rotation = self.rotation
            self._add_node(self.rbs)

        kwargs = self._render_kwargs.copy()
        kwargs['name'] = 'Mesh'
        kwargs['color'] = kwargs.get('color', (160 / 255, 160 / 255, 160 / 255, 1.0))
        self.mesh_seq = Meshes(self.vertices, self.faces, **kwargs)
        self.mesh_seq.position = self.position
        self.mesh_seq.rotation = self.rotation
        self._add_node(self.mesh_seq, gui_elements=['material'])

    @classmethod
    def from_amass(cls, npz_data_path, start_frame=None, end_frame=None, log=True, fps_out=None, **kwargs):
        """Load a sequence downloaded from the AMASS website."""
        body_data = np.load(npz_data_path)
        smpl_layer = SMPLLayer(model_type='smplh', gender=body_data['gender'].item(), device=C.device)

        if log:
            print('Data keys available: {}'.format(list(body_data.keys())))
            print('{:>6d} poses of size {:>4d}.'.format(body_data['poses'].shape[0], body_data['poses'].shape[1]))
            print('{:>6d} trans of size {:>4d}.'.format(body_data['trans'].shape[0], body_data['trans'].shape[1]))
            print('{:>6d} shape of size {:>4d}.'.format(1, body_data['betas'].shape[0]))
            print('Gender {}'.format(body_data['gender']))
            print('FPS {}'.format(body_data['mocap_framerate']))

        sf = start_frame or 0
        ef = end_frame or body_data['poses'].shape[0]
        poses = body_data['poses'][sf:ef]
        trans = body_data['trans'][sf:ef]

        if fps_out is not None:
            fps_in = body_data['mocap_framerate'].tolist()
            if fps_in != fps_out:
                ps = np.reshape(poses, [poses.shape[0], -1, 3])
                ps_new = resample_rotations(ps, fps_in, fps_out)
                poses = np.reshape(ps_new, [-1, poses.shape[1]])
                trans = resample_positions(trans, fps_in, fps_out)

        # Transform root orientation and translation into our viewer's coordinate system (where Y is up).
        to_y_up = np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])
        poses_root = R.as_matrix(R.from_rotvec(poses[:, 0:3]))
        poses_root = np.matmul(to_y_up[np.newaxis], poses_root)
        poses[:, 0:3] = R.as_rotvec(R.from_matrix(poses_root))
        trans = np.matmul(to_y_up[np.newaxis], trans[..., np.newaxis]).squeeze()
        i_root_end = 3
        i_body_end = i_root_end + smpl_layer.bm.NUM_BODY_JOINTS*3
        i_left_hand_end = i_body_end + smpl_layer.bm.NUM_HAND_JOINTS*3
        i_right_hand_end = i_left_hand_end + smpl_layer.bm.NUM_HAND_JOINTS*3

        return cls(poses_body=poses[:, i_root_end:i_body_end],
                   poses_root=poses[:, :i_root_end],
                   poses_left_hand=poses[:, i_body_end: i_left_hand_end],
                   poses_right_hand=poses[:, i_left_hand_end: i_right_hand_end],
                   smpl_layer=smpl_layer,
                   betas=body_data['betas'][np.newaxis],
                   trans=trans, **kwargs)

    @classmethod
    def from_3dpw(cls, pkl_data_path, smplx_neutral=False, **kwargs):
        """Load a 3DPW sequence which might contain multiple people."""
        with open(pkl_data_path, 'rb') as p:
            body_data = pkl.load(p, encoding='latin1')
        num_people = len(body_data['poses'])

        if smplx_neutral:
            smpl_layer = SMPLLayer(model_type='smplx', gender='neutral', num_betas=10, flat_hand_mean=True)

        poses_key = 'poses_smplx' if smplx_neutral else 'poses'
        trans_key = 'trans_smplx' if smplx_neutral else 'trans'
        betas_key = 'betas_smplx' if smplx_neutral else 'betas'

        name = kwargs.get('name', '3DPW')

        seqs = []
        for i in range(num_people):
            gender = body_data['genders'][i]
            if not smplx_neutral:
                smpl_layer = SMPLLayer(model_type='smpl', gender='female' if gender == 'f' else 'male', device=C.device,
                                       num_betas=10)

            # Extract the 30 Hz data that is already aligned with the image data.
            poses = body_data[poses_key][i]
            trans = body_data[trans_key][i]
            betas = body_data[betas_key][i]

            if len(betas.shape) == 1:
                betas = betas[np.newaxis]

            poses_body = poses[:, 3:]
            poses_root = poses[:, :3]
            trans_root = trans

            kwargs['name'] = name + " S{}".format(i)
            seq = cls(poses_body=poses_body, poses_root=poses_root, trans=trans_root,
                      smpl_layer=smpl_layer, betas=betas, **kwargs)
            seqs.append(seq)

        return seqs

    @classmethod
    def t_pose(cls, smpl_layer=None, betas=None, frames=1, **kwargs):
        """Creates a SMPL sequence whose single frame is a SMPL mesh in T-Pose."""

        if smpl_layer is None:
            smpl_layer = SMPLLayer(model_type='smplh', gender='neutral')

        poses = np.zeros([frames, smpl_layer.bm.NUM_BODY_JOINTS * 3])  # including hands and global root
        return cls(poses, smpl_layer, betas=betas, **kwargs)

    @property
    def vertex_normals(self):
        return self.mesh_seq.vertex_normals

    @property
    def poses(self):
        return torch.cat((self.poses_root, self.poses_body), dim=-1)

    def fk(self):
        """Get joints and/or vertices from the poses."""
        verts, joints = self.smpl_layer(poses_root=self.poses_root,
                                        poses_body=self.poses_body,
                                        poses_left_hand=self.poses_left_hand,
                                        poses_right_hand=self.poses_right_hand,
                                        betas=self.betas,
                                        trans=self.trans)
        skeleton = self.smpl_layer.skeletons()['body'].T
        faces = self.smpl_layer.bm.faces.astype(np.int64)
        joints = joints[:, :skeleton.shape[0]]
        return c2c(verts), c2c(joints), c2c(faces), c2c(skeleton)

    def interpolate(self, frame_ids):
        """
        Replace the frames at the given frame IDs via an interpolation of its neighbors. Only the body pose as well
        as the root pose and translation are interpolated.
        :param frame_ids: A list of frame ids to be interpolated.
        """
        ids = np.unique(frame_ids)
        all_ids = np.arange(self.n_frames)
        mask_avail = np.ones(self.n_frames, dtype=np.bool)
        mask_avail[ids] = False

        # Interpolate poses.
        all_poses = torch.cat([self.poses_root, self.poses_body], dim=-1)
        ps = np.reshape(all_poses.cpu().numpy(), (self.n_frames, -1, 3))
        ps_interp = interpolate_rotations(ps[mask_avail], all_ids[mask_avail], ids)
        all_poses[ids] = torch.from_numpy(ps_interp.reshape(len(ids), -1)).to(dtype=self.betas.dtype,
                                                                              device=self.betas.device)
        self.poses_root = all_poses[:, :3]
        self.poses_body = all_poses[:, 3:]

        # Interpolate global translation.
        ts = self.trans.cpu().numpy()
        ts_interp = interpolate_positions(ts[mask_avail], all_ids[mask_avail], ids)
        self.trans[ids] = torch.from_numpy(ts_interp).to(dtype=self.betas.dtype, device=self.betas.device)

        self.redraw()

    def redraw(self):
        self.vertices, self.joints, self.faces, self.skeleton = self.fk()
        if self._is_rigged:
            self.skeleton_seq.joint_positions = self.joints
        self.mesh_seq.vertices = self.vertices
        super().redraw()

    def gui(self, imgui):
        super().gui_animation(imgui)
        super().gui_position(imgui)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import torch

from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig


class Configuration(object):

    class __Configuration:

        def next_gui_id(self):
            self._gui_counter += 1
            return self._gui_counter

        def __init__(self):
            # Load the default configurations.
            self._conf = OmegaConf.load(os.path.join(os.path.dirname(__file__), 'aitvconfig.yaml'))

            # Check if we have a config file in AITVRC env variable.
            if os.environ.get('AITVRC', None) is not None:
                env_conf = os.environ['AITVRC']
                if os.path.isdir(env_conf):
                    conf = OmegaConf.load(os.path.join(env_conf, 'aitvconfig.yaml'))
                else:
                    conf = OmegaConf.load(env_conf)
                self._conf.merge_with(conf)

            # Check if we have a config file in the working directory which overrides all previous configs.
            local_conf = os.path.join(os.getcwd(), 'aitvconfig.yaml')
            if os.path.exists(local_conf):
                conf = OmegaConf.load(local_conf)
                self._conf.merge_with(conf)

            self._gui_counter = 0
            self._gpu_available = torch.cuda.is_available()

        def update_conf(self, conf_obj):
            """Update the configuration with another configuration file or another OmegaConf configuration object."""
            if isinstance(conf_obj, str):
                conf_obj = OmegaConf.load(conf_obj)
            else:
                assert isinstance(conf_obj, DictConfig)
            self._conf.merge_with(conf_obj)

        def __getattr__(self, item):
            if hasattr(self._conf, item):
                # Some attributes of the config are converted to torch objects automatically.
                if item == 'device':
                    return torch.device(self._conf.get('device', 'cuda:0') if self._gpu_available else 'cpu')
                elif item == 'f_precision':
                    return getattr(torch, 'float{}'.format(self._conf.get('f_precision', 32)))
                elif item == 'i_precision':
                    return getattr(torch, 'int{}'.format(self._conf.get('i_precision', 64)))
                else:
                    return getattr(self._conf, item)
            else:
                # Default behavior.
                return self.__getattribute__(item)

    instance = None

    def __new__(cls, *args, **kwargs):
        if not Configuration.instance:
            Configuration.instance = Configuration.__Configuration()
        return Configuration.instance

    def __getattr__(self, item):
        return getattr(self.instance, item)

    def __setattr__(self, key, value):
        return setattr(self.instance, key, value)


CONFIG = Configuration()



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.scene.node import Node
from aitviewer.scene.camera_utils import look_at
from aitviewer.scene.camera_utils import orthographic_projection


class Light(Node):
    """Simple point light."""

    def __init__(self, intensity_diffuse=1.0, intensity_ambient=1.0, **kwargs):
        super(Light, self).__init__(**kwargs)

        self.intensity_ambient = intensity_ambient
        self.intensity_diffuse = intensity_diffuse
        self.shadow_map_size = 30.0

    def mvp(self):
        """Return a model-view-projection matrix to project vertices into the view of the light."""
        s = self.shadow_map_size
        P = orthographic_projection(s, s, 0.01, 100.0)
        V = look_at(np.array(self.position), np.array([0.0, 0.0, 0.0]), np.array([0.0, 1.0, 0.0]))
        return (P @ V).astype('f4')

    def gui(self, imgui):
        self.gui_position(imgui)
        self.gui_material(imgui, show_advanced=False)
        # Position control
        # _, self.position = imgui.drag_float3('Position##pos', *self.position, 0.1, format='%.2f')


        # Light controls
        _, self.intensity_ambient = imgui.drag_float('Ambient##ambient', self.intensity_ambient, 0.01, min_value=0.0, max_value=1.0,
                                           format='%.2f')
        _, self.intensity_diffuse = imgui.drag_float('Diffuse##diffuse', self.intensity_diffuse, 0.01, min_value=0.0, max_value=1.0,
                                           format='%.2f')

        # _, self.shadow_map_size = imgui.drag_float('Shadow Map Size##s', self.shadow_map_size, 0.1, format='%.2f', min_value=0.01, max_value=100.0)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import tempfile
import shutil

from aitviewer.scene.camera import PinholeCamera
from aitviewer.utils import images_to_video
from aitviewer.viewer import Viewer
from tqdm import tqdm


class HeadlessRenderer(Viewer):
    gl_version = (3, 3)
    samples = 0  # Headless rendering does not like super sampling.
    window_type = 'headless'

    def __init__(self, **kwargs):
        """
        Initializer.
        :param frame_dir: Where to save the frames to.
        :param kwargs: kwargs.
        """
        super().__init__(**kwargs)

        # Scene setup.
        self.camera = PinholeCamera(45.0)
        self.draw_edges = False

        # Book-keeping for the headless rendering.
        self.n_frames_rendered = 0
        self.frame_dir = None
        self.progress_bar = None

    def run(self, frame_dir=None, video_dir=None, keep_frames=False, log=True, load_cam=False):
        """
        Convenience method to run the headless rendering.
        :param frame_dir: Where to store the individual frames or None if you don't care.
        :param video_dir: If set will automatically generate a video from the images found in `frame_dir`. Must
          be specified if `frame_dir` is None.
        :param keep_frames: Whether to keep the individual frames or automatically delete them. This is ignored if
         `frame_dir` is set, i.e. when `frame_dir` is set, we never delete frames.
        :param log: Log some info.
        """
        if frame_dir is None and video_dir is None:
            raise ValueError("You should either specify a path where to render the images to or where to "
                             "save the video to.")

        if video_dir is not None:
            assert video_dir.endswith(".mp4")

        if frame_dir is None:
            temp_dir = tempfile.TemporaryDirectory()
            frame_dir = tempfile.TemporaryDirectory().name
        else:
            temp_dir = None
            # We want to avoid overriding anything in an existing directory, so add suffixes.
            format_str = "{:0>4}"
            counter = 0
            candidate_dir = os.path.join(frame_dir, format_str.format(counter))
            while os.path.exists(candidate_dir):
                counter += 1
                candidate_dir = os.path.join(frame_dir, format_str.format(counter))
            frame_dir = os.path.abspath(candidate_dir)

        # The frame dir does not yet exist (we've made sure of it).
        os.makedirs(frame_dir)
        self.frame_dir = frame_dir

        self.scene.make_renderable(self.ctx)
        if self.auto_set_floor:
            self.scene.auto_set_floor()

        if self.auto_set_camera_target:
            self.scene.auto_set_camera_target()

        if load_cam:
            self.scene.camera.load_cam()

        self.timer.start()
        while not self.rendering_finished():
            current_time, delta = self.timer.next_frame()
            self.window.clear()
            self.window.render(current_time, delta)
            self.window.swap_buffers()
            self.scene.next_frame()
        _, duration = self.timer.stop()

        if duration > 0 and log:
            print("Duration: {0:.2f}s @ {1:.2f} FPS".format(duration, self.max_frame / duration))

        if video_dir is not None:
            images_to_video(frame_dir, video_dir)

        # Only delete the frames if it was a temporary directory and we don't want to keep them.
        if temp_dir is not None and not keep_frames:
            shutil.rmtree(frame_dir)

    @property
    def max_frame(self):
        return self.scene.n_frames

    def rendering_finished(self):
        return not self.n_frames_rendered < self.max_frame

    def render(self, time, frame_time):
        self.render_shadowmap()
        self.render_prepare()
        self.render_scene()

        self.n_frames_rendered += 1
        if self.progress_bar is None:
            self.progress_bar = tqdm(total=self.max_frame, desc='Rendering frames')

        if self.n_frames_rendered > self.max_frame:
            self.progress_bar.close()
            self.wnd.close()
        else:
            self.progress_bar.update()
            self.save_current_frame_as_image(self.frame_dir, self.n_frames_rendered - 1)


# def _instantiate_and_run_window(window_cls: Viewer, *args, log=True):
#     """
#     Instantiate the window passing the provided args into the constructor and enter a blocking visualization loop.
#     This is built following `moderngl_window.run_window_config`.
#
#     :param window_cls: The window to run.
#     :param args: The arguments passed to `conig_cls` constructor.
#     :param log: Whether to log to the console.
#     """
#     base_window_cls = get_local_window_cls(window_cls.window_type)
#
#     # Calculate window size
#     size = window_cls.window_size
#     size = int(size[0] * window_cls.size_mult), int(size[1] * window_cls.size_mult)
#
#     window = base_window_cls(
#         title=window_cls.title,
#         size=size,
#         fullscreen=window_cls.fullscreen,
#         resizable=window_cls.resizable,
#         gl_version=window_cls.gl_version,
#         aspect_ratio=window_cls.aspect_ratio,
#         vsync=window_cls.vsync,
#         samples=window_cls.samples,
#         cursor=False,
#     )
#     window.print_context_info()
#     activate_context(window=window)
#     timer = Timer()
#     window.config = window_cls(*args, ctx=window.ctx, wnd=window, timer=timer)
#
#     timer.start()
#
#     while not window.is_closing:
#         current_time, delta = timer.next_frame()
#         window.clear()
#         window.render(current_time, delta)
#         window.swap_buffers()
#
#     _, duration = timer.stop()
#     window.destroy()
#     if duration > 0 and log:
#         print("Duration: {0:.2f}s @ {1:.2f} FPS".format(duration, window.frames / duration))



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

def requires_ctx(func):
    def _decorator(self, *args, **kwargs):
        if self.ctx is None:
            raise ValueError(
                'This function needs access to an OpenGL context - did you forget to call make_renderable()?')
        else:
            return func(self, *args, **kwargs)
    return _decorator


def default_to_current_frame(func):
    def _decorator(self, *args, **kwargs):
        if len(args) == 0:
            kwargs['frame_id'] = kwargs.get('frame_id', self._current_frame_id)
        return func(self, *args, **kwargs)
    return _decorator


def hooked(func):
    """Decarotor to call the super method of a function before calling the function itself."""
    def _decorator(self, *args, **kwargs):
        super_obj = super(type(self), self)
        super_fn = getattr(super_obj, func.__name__)
        super_fn(*args, **kwargs)
        return func(self, *args, **kwargs)
    return _decorator



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from aitviewer.scene.node import Node


class Streamable(Node):
    """Interface for renderables."""

    def __init__(self, **kwargs):
        super(Streamable, self).__init__(**kwargs)

        self.is_recording = False

    def start(self):
        pass

    def stop(self):
        pass

    def capture(self):
        """Capture from the sensor"""
        raise NotImplementedError("Must be implemented by the subclass.")

    def record_start(self):
        self.is_recording = True

    def record_capture(self):
        pass

    def record_finish(self):
        self.is_recording = False
        return []




"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from moderngl_window import resources
from moderngl_window.meta import ProgramDescription


def _load(name):
    return resources.programs.load(ProgramDescription(path=name))


def get_simple_unlit_program():
    return _load('simple_unlit.glsl')


def get_smooth_lit_with_edges_program():
    return _load('smooth_lit_with_edges.glsl')


def get_flat_lit_with_edges_program():
    return _load('flat_lit_with_edges.glsl')


def get_cylinder_program():
    return _load('cylinder.glsl')


def get_screen_texture_program():
    return _load('screen_texture.glsl')


def get_smooth_lit_texturized_program():
    return _load('smooth_lit_tex.glsl')



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import os
from aitviewer.scene.node import Node
from aitviewer.shaders import get_simple_unlit_program
from moderngl_window.opengl.vao import VAO
import moderngl


class PointClouds(Node):
    """
    Draw a point clouds man!
    """

    def __init__(self, points, colors=None, lengths=None, point_size=5.0, color=(0.0, 0.0, 1.0, 1.0), **kwargs):
        """
        A sequence of point clouds. Each point cloud can have a varying number of points.
        Internally represented as a list of arrays.
        :param points: Sequence of points (F, P, 3)
        :param colors: Sequence of Colors (F, C, 4)
        :param lengths: Length mask for each frame of points denoting the usable part of the array
        :param point_size: Initial point size
        """
        self.points = points
        super(PointClouds, self).__init__(n_frames=len(self.points), color=color, **kwargs)

        self.n_points = self.points.shape[1]
        self.colors = colors
        self.point_size = point_size
        self.lengths = lengths

        self.vao = VAO("points", mode=moderngl.POINTS)

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, points):
        if len(points.shape) == 2 and points.shape[-1] == 3:
            points = points[np.newaxis]
        assert len(points.shape) == 3

        self._points = points
        self.n_frames = len(points)

    @property
    def colors(self):
        return self._colors

    @colors.setter
    def colors(self, colors):
        # Colors cannot be empty
        if colors is None:
            self._colors = np.full((self.n_frames, self.n_points, 4), self.color)
        elif isinstance(colors, tuple) and len(colors) == 4:
            self._vertex_colors = np.full((self.n_frames, self.n_points, 4), colors)
        else:
            # Ensure batch dimension is in place
            if len(colors.shape) == 2 and colors.shape[-1] == 4:
                colors = colors[np.newaxis]
            assert len(colors.shape) == 3

            self._colors = colors

    @Node.color.setter
    def color(self, color):
        alpha_changed = np.abs((np.array(color) - np.array(self._color)))[-1] > 0
        self.material.color = color
        if self.is_renderable:
            # If alpha changed, don't update all colors
            if alpha_changed:
                self._colors[..., -1] = color[-1]
            else:
                self.colors = color

        self.redraw()

    @property
    def current_points(self):
        return self.points[self.current_frame_id]

    @property
    def current_colors(self):
        return self.colors[self.current_frame_id]

    @property
    def bounds(self):
        return self.get_bounds(self.points)

    def on_frame_update(self):
        """Called whenever a new frame must be displayed."""
        super().on_frame_update()
        self.redraw()

    def redraw(self):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable:
            return

        # Each write call takes about 1-2 ms
        points = self.current_points
        colors = self.current_colors

        self._clear_buffer()
        self.vbo_points.write(points.astype('f4').tobytes())
        self.vbo_colors.write(colors.astype('f4').tobytes())

    def _clear_buffer(self):
        self.vbo_points.clear()
        self.vbo_colors.clear()

    @Node.once
    def make_renderable(self, ctx):
        ctx.point_size = self.point_size

        self.prog = get_simple_unlit_program()
        self.vbo_points = ctx.buffer(self.current_points.astype('f4').tobytes())
        self.vbo_colors = ctx.buffer(self.current_colors.astype('f4').tobytes())
        self.vao.buffer(self.vbo_points, '3f', ['in_position'])
        self.vao.buffer(self.vbo_colors, '4f', ['in_color'])

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        self.vao.render(self.prog)

    def to_npz(self, path=''):
        np.savez_compressed(path + self.name + '.npz',
                            points=self.points,
                            colors=self.colors,
                            lengths=self.lengths)

    @classmethod
    def from_npz(cls, npz_path):
        """Load a sequence from an npz file. The filename becomes the name of the sequence"""
        data = np.load(npz_path, allow_pickle=True)
        name = os.path.splitext(os.path.basename(npz_path))[0]
        return cls(
            points= data['points'] if 'points' in data and data['points'].size > 1 else None,
            colors=data['colors'] if 'colors' in data and data['colors'].size > 1 else None,
            lengths=data['lengths'] if 'lengths' in data and data['lengths'].dtype != object else None
        )

    def gui(self, imgui):
        super().gui(imgui)
        if imgui.button("Save"):
            self.to_npz()


"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.scene.material import Material
from aitviewer.scene.node import Node
from aitviewer.renderables.spheres import Spheres
from aitviewer.renderables.lines import Lines


class Skeletons(Node):
    """
    Render a skeleton as a set of spheres that are connected with cone-shaped lines.
    """
    def __init__(self,
                 joint_positions,
                 joint_connections,
                 joint_rotations=None,
                 radius=0.01,
                 color=(1.0, 177 / 255, 1 / 255, 1.0),
                 **kwargs):
        """
        Initializer.
        :param joint_positions: A np array of shape (F, J, 3) containing J joint positions over F many time steps.
        :param joint_connections: The definition of the skeleton as a numpy array of shape (N_LINES, 2) where each row
          defines one connection between joints. The max entry in this array must be < J.
        :param joint_rotations: Optional joint orientations as a np array of shape (F, J, 3, 3).
        :param radius: Radius of the sphere located at each joint's position.
        :param color: 4-tuple color, yellow by default.
        :param kwargs: Remaining render arguments.
        """
        if not isinstance(joint_connections, np.ndarray):
            joint_connections = np.array(joint_connections)

        self.joint_positions = joint_positions
        super(Skeletons, self).__init__(n_frames=self.joint_positions.shape[0], color=color, **kwargs)

        self.joint_rotations = joint_rotations
        self.joint_connections = joint_connections

        # Root nodes are not connected to any other joints, so ignore those.
        non_root = ~np.any(joint_connections == -1, axis=-1)
        self.skeleton = joint_connections[non_root]
        self.joint_idxs = np.unique(self.skeleton)

        # Nodes.
        material = Material(color=color)
        self.spheres = Spheres(joint_positions, radius=radius, material=material)
        self.lines = Lines(lines=self.joint_positions[:, self.skeleton].reshape(len(self), -1, 3),
                           mode='lines', r_base=radius, r_tip=radius / 10.0,material=material)
        self._add_nodes(self.spheres, self.lines, show_in_hierarchy=False)

    @property
    def joint_positions(self):
        return self._joint_positions

    @joint_positions.setter
    def joint_positions(self, joint_positions):
        if len(joint_positions.shape) == 2 and joint_positions.shape[-1] == 3:
            joint_positions = joint_positions[np.newaxis]
        assert len(joint_positions.shape) == 3
        self._joint_positions = joint_positions
        self.n_frames = len(joint_positions)

    def redraw(self):
        self.spheres.sphere_positions = self.joint_positions
        self.lines.lines = self.joint_positions[:, self.skeleton].reshape(len(self), -1, 3)
        super().redraw()

    @property
    def bounds(self):
        return self.get_bounds(self.joint_positions)

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.spheres.color = color
        self.lines.color = color



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np

from aitviewer.renderables.meshes import Meshes
from aitviewer.scene.node import Node
from aitviewer.shaders import get_smooth_lit_with_edges_program
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties


class Plane(Node):
    """
    Draw a plane.
    """
    def __init__(self,
                 center,
                 v1,
                 v2,
                 size=10.0,
                 color=(0.5, 0.5, 0.5, 1.0),
                 **kwargs):
        """
        Initializer.
        :param center: Center of the plane.
        :param v1: 3D vector lying in the plane.
        :param v2: 3D vector lying in the plane and not co-linear to `v2`.
        :param size: Size of the plane.
        :param color: Color of the plane.
        """
        super(Plane, self).__init__(color=color, **kwargs)
        if np.dot(v1, v2) > 0.00001:
            raise ValueError('v1 and v2 are not orthogonal.')

        self.center = center
        self.v1 = v1 / np.linalg.norm(v1)
        self.v2 = v2 / np.linalg.norm(v2)
        self.size = size

        self.vertices, self.normals = self._get_renderable_data()
        self.colors = np.full((self.vertices.shape[0], 4), color)

    def _get_renderable_data(self):
        p0 = self.center + self.v1 * self.size - self.v2 * self.size
        p1 = self.center + self.v1 * self.size + self.v2 * self.size
        p2 = self.center - self.v1 * self.size + self.v2 * self.size
        p3 = self.center - self.v1 * self.size - self.v2 * self.size
        normal = np.cross(self.v2, self.v1)
        normal = np.tile(normal[np.newaxis], (4, 1))
        return np.row_stack([p1, p0, p2, p3]), normal

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.colors = np.full((self.vertices.shape[0], 4), color)
        if self.is_renderable:
            self.vbo_colors.write(self.colors.astype('f4').tobytes())

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        self.prog = get_smooth_lit_with_edges_program()
        self.vbo_vertices = ctx.buffer(self.vertices.astype('f4').tobytes())
        self.vbo_normals = ctx.buffer(self.normals.astype('f4').tobytes())
        self.vbo_colors = ctx.buffer(self.colors.astype('f4').tobytes())

        self.vao = ctx.vertex_array(self.prog,
                                    [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                     (self.vbo_normals, '3f4 /v', 'in_normal'),
                                     (self.vbo_colors, '4f4 /v', 'in_color')])

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        self.receive_shadow(self.prog, **kwargs)

        self.prog['draw_edges'].value = 1.0 if kwargs['draw_edges'] else 0.0
        self.prog['norm_coloring'].value = False
        self.prog['win_size'].value = kwargs['window_size']

        set_lights_in_program(self.prog, kwargs['lights'])
        set_material_properties(self.prog, self.material)
        self.vao.render(moderngl.TRIANGLE_STRIP)


class Chessboard(Node):
    """A plane that is textured like a chessboard."""
    def __init__(self,
                 side_length,
                 n_tiles,
                 color1=(0.0, 0.0, 0.0, 1.0),
                 color2=(1.0, 1.0, 1.0, 1.0),
                 plane='xz',
                 height=0.0,
                 tiling=True,
                 **kwargs):
        """
        Initializer.
        :param side_length: Length of one side of the plane.
        :param n_tiles: Number of tiles for the chessboard pattern.
        :param color1: First color of the chessboard pattern.
        :param color2: Second color of the chessboard pattern.
        :param plane: In which plane the chessboard lies. Allowed are 'xz', 'xy', 'yz'.
        :param height: The height of the plane.
        :param kwargs: Remaining kwargs.
        """
        assert plane in ['xz', 'xy', 'yz']
        super(Chessboard, self).__init__(**kwargs)
        self.side_length = side_length
        self.n_tiles = n_tiles
        self.c1 = np.array(color1)
        self.c2 = np.array(color2)
        self.plane = plane
        self.height = height
        self.tiling = tiling

        vs, fs, fc, c1_idxs, c2_idxs = self._construct_board()
        self.fcs_tiled = fc
        self.c1_idxs = c1_idxs
        self.c2_idxs = c2_idxs

        self.mesh = Meshes(vs, fs, face_colors=fc)
        self.mesh.position = self.position
        self.add(self.mesh, has_gui=False, show_in_hierarchy=False)

    # noinspection PyAttributeOutsideInit
    def _construct_board(self):
        """Construct the chessboard mesh."""
        vertices = []
        faces = []
        face_colors = []
        c1_idxs = []  # Store indices into face_colors containing color 1.
        c2_idxs = []  # Store indices into face_colors containing color 2.
        tl = self.side_length / self.n_tiles
        dim1 = 'xyz'.index(self.plane[0])
        dim2 = 'xyz'.index(self.plane[1])
        up = 'xyz'.index('xyz'.replace(self.plane[0], '').replace(self.plane[1], ''))

        for r in range(self.n_tiles):
            for c in range(self.n_tiles):
                v0 = np.zeros([3])
                v0[dim1] = r * tl
                v0[dim2] = c * tl

                v1 = np.zeros([3])
                v1[dim1] = (r + 1) * tl
                v1[dim2] = c * tl

                v2 = np.zeros([3])
                v2[dim1] = (r + 1) * tl
                v2[dim2] = (c + 1) * tl

                v3 = np.zeros([3])
                v3[dim1] = r * tl
                v3[dim2] = (c + 1) * tl

                vertices.extend([v0, v1, v2, v3])

                # Need counter-clock-wise ordering
                faces.append([len(vertices) - 4, len(vertices) - 1, len(vertices) - 3])
                faces.append([len(vertices) - 3, len(vertices) - 1, len(vertices) - 2])

                if r % 2 == 0 and c % 2 == 0:
                    c = self.c1
                    fc_idxs = c1_idxs
                elif r % 2 == 0 and c % 2 != 0:
                    c = self.c2
                    fc_idxs = c2_idxs
                elif r % 2 != 0 and c % 2 != 0:
                    c = self.c1
                    fc_idxs = c1_idxs
                else:
                    c = self.c2
                    fc_idxs = c2_idxs
                face_colors.append(c)
                face_colors.append(c)
                fc_idxs.extend([len(face_colors) - 2, len(face_colors) - 1])

        vs = np.stack(vertices)
        vs = vs - np.mean(vertices, axis=0, keepdims=True)
        vs[:, up] = self.height
        fs = np.stack(faces)
        cs = np.stack(face_colors)

        return vs, fs, cs, c1_idxs, c2_idxs

    def _update_colors(self):
        self.fcs_tiled[self.c1_idxs] = self.c1
        self.fcs_tiled[self.c2_idxs] = self.c2 if self.tiling else self.c1
        self.mesh.face_colors = self.fcs_tiled

    def gui(self, imgui):
        self.gui_position(imgui)

        u, c1 = imgui.color_edit4("Color 1##color{}'".format(self.unique_name), *self.c1, show_alpha=True)
        if u:
            self.c1 = c1
            self._update_colors()

        u, c2 = imgui.color_edit4("Color 2##color{}'".format(self.unique_name), *self.c2, show_alpha=True)
        if u:
            self.c2 = c2
            self._update_colors()

        u, self.tiling = imgui.checkbox('Toggle Tiling', self.tiling)
        if u:
            self._update_colors()




"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class ExceptionModule(object):
    """
    Source: Trimesh package @ https://github.com/mikedh/trimesh/blob/master/trimesh/exceptions.py

    Create a dummy module which will raise an exception when attributes
    are accessed.
    For soft dependencies we want to survive failing to import but
    we would like to raise an appropriate error when the functionality is
    actually requested so the user gets an easily debuggable message.
    """

    def __init__(self, exc):
        self.exc = exc

    def __getattribute__(self, *args, **kwargs):
        # if it's asking for our class type return None
        # this allows isinstance() checks to not re-raise
        if args[0] == '__class__':
            return None.__class__
        # otherwise raise our original exception
        raise super(ExceptionModule, self).__getattribute__('exc')




"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.configuration import CONFIG as C
from aitviewer.scene.material import Material
from moderngl_window.opengl.vao import VAO


class Node(object):
    """Interface for nodes."""

    def __init__(self,
                 name=None,
                 icon=None,
                 position=None,
                 rotation=None,
                 scale=1.0,
                 color=(0.5, 0.5, 0.5, 1.0),
                 material=None,
                 render_priority=1,
                 n_frames=1):
        """
        :param name: Name of the node
        :param icon: Custom Node Icon using custom Icon font
        :param position: Starting position in the format (X,Y,Z)
        :param rotation: Starting rotation in rotation matrix representation (3,3)
        :param scale: Starting scale (scalar)
        :param color: (R,G,B,A) 0-1 formatted color value.
        :param material: Object material properties. The color specified in the material will override node color
        :param n_frames: How many frames this renderable has.
        """

        # Transform & Animation
        self._position = np.array([0.0, 0.0, 0.0]) if position is None else position
        self._rotation = np.eye(3) if rotation is None else rotation
        self._scale = scale
        self._n_frames = n_frames
        self._current_frame_id = 0

        # Material
        self.material = Material(color=color) if material is None else material

        # Renderable Attributes
        self.render_priority = render_priority
        self.is_renderable = False
        # To render shadows we need a separate VAO.
        self._shadow_vao = None
        self._fragmap_vao = None

        # GUI
        self.name = name if name is not None else type(self).__name__
        self.uid = C.next_gui_id()
        self.unique_name = self.name + "{}".format(self.uid)
        self.icon = icon if icon is not None else '\u0082'
        self._enabled = True
        self._expanded = False
        self._has_gui = True
        self._gui_elements = []
        self._show_in_hierarchy = True

        self.nodes = []

    # Transform
    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._position = position
        for n in self.nodes:
            n.position = position

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, rotation):
        self._rotation = rotation
        for n in self.nodes:
            n.rotation = rotation

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    def model_matrix(self):
        """Construct model matrix from this node's orientation and position."""
        rotation = np.eye(4)
        rotation[:3, :3] = self.rotation
        trans = np.eye(4)
        trans[:3, 3] = self.position
        scale = np.eye(4)
        scale[([0, 1, 2], [0, 1, 2])] = self.scale
        return (rotation @ trans @ scale).astype('f4')

    @property
    def color(self):
        return self.material.color

    @color.setter
    def color(self, color):
        self.material.color = color

    @property
    def bounds(self):
        """ The bounds in the format ((x_min, x_max), (y_min, y_max), (z_min, z_max)) """
        return None

    def get_bounds(self, points):
        if len(points.shape) == 2 and points.shape[-1] == 3:
            points = points[np.newaxis]
        assert len(points.shape) == 3
        return np.array([
            [points[:, :, 0].min(), points[:, :, 0].max()],
            [points[:, :, 1].min(), points[:, :, 1].max()],
            [points[:, :, 2].min(), points[:, :, 2].max()]])

    @property
    def n_frames(self):
        return self._n_frames

    @n_frames.setter
    def n_frames(self, n_frames):
        self._n_frames = n_frames

    def __len__(self):
        return self.n_frames

    @property
    def current_frame_id(self):
        return self._current_frame_id

    @current_frame_id.setter
    def current_frame_id(self, frame_id):
        if self.n_frames == 1 or frame_id == self._current_frame_id:
            return
        if frame_id < 0:
            self._current_frame_id = 0
        elif frame_id >= len(self):
            self._current_frame_id = len(self) - 1
        else:
            self._current_frame_id = frame_id
        self.on_frame_update()

    def next_frame(self):
        self.current_frame_id = self.current_frame_id + 1 if self.current_frame_id < len(self) - 1 else 0

    def previous_frame(self):
        self.current_frame_id = self.current_frame_id - 1 if self.current_frame_id > 0 else len(self) - 1

    def on_frame_update(self):
        for n in self.nodes:
            n.current_frame_id = self.current_frame_id

    def add(self, *nodes, **kwargs):
        self._add_nodes(*nodes, **kwargs)

    def _add_node(self,
                  n,
                  has_gui=True,
                  gui_elements=None,
                  show_in_hierarchy=True,
                  expanded=False,
                  enabled=True):
        """
        Add a single node
        :param has_gui: Whether the node has a GUI.
        :param gui_elements: Which elements of the GUI are displayed by default.
        :param show_in_hierarchy: Whether to show the node in the scene hierarchy.
        :param expanded: Whether the node is initially expanded in the GUI.
        """
        if n is None:
            return
        n._has_gui = has_gui
        n._gui_elements = gui_elements if gui_elements is not None else ['animation', 'position', 'material']
        n._show_in_hierarchy = show_in_hierarchy
        n._expanded = expanded
        n._enabled = enabled
        self.nodes.append(n)

    def _add_nodes(self, *nodes, **kwargs):
        """Add multiple nodes"""
        for n in nodes:
            self._add_node(n, **kwargs)

    @property
    def has_gui(self):
        return self._has_gui

    @property
    def show_in_hierarchy(self):
        return self._show_in_hierarchy

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        self._enabled = enabled
        for n in self.nodes:
            n.enabled = enabled

    @property
    def expanded(self):
        return self._expanded

    @expanded.setter
    def expanded(self, expanded):
        self._expanded = expanded

    def gui_animation(self, imgui):
        # Animation Control
        if self.n_frames > 1 and 'animation' in self._gui_elements:
            u, fid = imgui.drag_int('Frame##r_{}'.format(self.unique_name),
                                    self.current_frame_id, min_value=0, max_value=self.n_frames - 1)
            if u:
                self.current_frame_id = fid

    def gui_position(self, imgui):
        # Position controls
        u, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 0.1, format='%.2f')
        if u:
            self.position = pos

    def gui_scale(self, imgui):
        # Scale controls
        u, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.scale, 0.01, min_value=0.001,
                                    max_value=10.0, format='%.3f')
        if u:
            self.scale = scale

    def gui_material(self, imgui, show_advanced=True):
        # Color Control
        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color, show_alpha=True)
        if uc:
            self.color = color

        if show_advanced:
            if imgui.tree_node("Advanced material##advanced_material{}'".format(self.unique_name)):
                # Diffuse
                ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name), self.material.diffuse,  0.0, 1.0, '%.2f')
                if ud:
                    self.material.diffuse = diffuse

                # Ambient
                ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name), self.material.ambient,  0.0, 1.0, '%.2f')
                if ua:
                    self.material.ambient = ambient

                imgui.tree_pop()

    def gui(self, imgui):
        """
        Render the GUI that gets displayed in the scene window. Implementation optional.
        Elements rendered here will show up in the scene hierarchy
        :param imgui: imgui context.
        See https://pyimgui.readthedocs.io/en/latest/reference/imgui.core.html for available elements to render
        """
        if 'animation' in self._gui_elements:
            self.gui_animation(imgui)

        if 'position' in self._gui_elements:
            self.gui_position(imgui)

        if 'material' in self._gui_elements:
            self.gui_material(imgui)

        if 'scale' in self._gui_elements:
            self.gui_scale(imgui)

    # Renderable
    @staticmethod
    def once(func):
        def _decorator(self, *args, **kwargs):
            if self.is_renderable:
                return
            else:
                func(self, *args, **kwargs)
                self.is_renderable = True
        return _decorator

    def make_renderable(self, ctx):
        """
        Prepares this object for rendering. This function must be called before `render` is used.
        :param ctx: The moderngl context.
        """
        pass

    def render(self, camera, position=None, rotation=None, **kwargs):
        """Render the current frame in this sequence."""
        pass

    def redraw(self, **kwargs):
        """ Perform update and redraw operations. Push to the GPU when finished. Recursively redraw child nodes"""
        for n in self.nodes:
            n.redraw()

    def set_camera_matrices(self, prog, camera, **kwargs):
        """Set the model view projection matrix in the given program."""
        w, h = kwargs['window_size'][0], kwargs['window_size'][1]
        mvp = np.matmul(camera.get_view_projection_matrix(w, h), self.model_matrix())
        # Transpose because np is row-major but OpenGL expects column-major.
        prog['mvp'].write(mvp.T.astype('f4').tobytes())

    def make_pickable(self, vertex_buffer, index_buffer=None):
        """
        Call this function if the renderable can be clicked on.
        :param vertex_buffer: All the vertex data that makes up this renderable. Can be a ModernGL VBO or a numpy array.
        :param index_buffer: Optional index buffer if this renderable uses indexed rendering.
        """
        if self._fragmap_vao is None:
            self._fragmap_vao = VAO('{}:fragmap'.format(self.unique_name))
            self._fragmap_vao.buffer(vertex_buffer, '3f', ['in_position'])
            if index_buffer is not None:
                self._fragmap_vao.index_buffer(index_buffer)
        else:
            raise ValueError('Fragment map VAO already created.')

    def cast_shadow(self, vertex_buffer, index_buffer=None):
        """
        Call this function if the renderable is to cast a shadow.
        :param vertex_buffer: All the vertex data that makes up this renderable. Can be a ModernGL VBO or a numpy array.
        :param index_buffer: Optional index buffer if this renderable uses indexed rendering.
        """
        if self._shadow_vao is None:
            self._shadow_vao = VAO('{}:shadow'.format(self.unique_name))
            self._shadow_vao.buffer(vertex_buffer, '3f', ['in_position'])
            if index_buffer is not None:
                self._shadow_vao.index_buffer(index_buffer)
        else:
            raise ValueError('Shadow VAO already created.')

    def receive_shadow(self, program, **kwargs):
        """
        Call this function if the renderable is to receive shadows.
        :param program: The shader program that can shade with shadows.
        :param kwargs: The render kwargs.
        """
        if kwargs.get('shadows_enabled', False):
            light_mvp = kwargs['light_mvp'] @ self.model_matrix()
            program['light_mvp'].write(light_mvp.T.tobytes())
            kwargs['offscreen_depth'].use(location=0)

    def render_shadowmap(self, light_mvp, program):
        if self._shadow_vao is None:
            return

        mvp = light_mvp @ self.model_matrix()
        program['mvp'].write(mvp.T.tobytes())

        self._shadow_vao.render(program)

    def render_fragmap(self, camera, prog, window_size):
        if self._fragmap_vao is None:
            return

        p = camera.get_projection_matrix(window_size[0], window_size[1])
        mv = camera.view() @ self.model_matrix()

        # Transpose because np is row-major but OpenGL expects column-major.
        prog['projection'].write(p.T.astype('f4').tobytes())
        prog['modelview'].write(mv.T.astype('f4').tobytes())
        prog['obj_id'] = self.uid

        self._fragmap_vao.render(prog)

    def release(self):
        """Release all resources occupied by this node."""
        pass



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import torch
import os
import subprocess

from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa
from scipy.interpolate import CubicSpline


def to_torch(x, dtype, device):
    if x is None:
        return None
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).to(dtype=dtype, device=device)
    return x.to(dtype=dtype, device=device)


def to_numpy(x):
    if isinstance(x, np.ndarray):
        return x
    return x.detach().cpu().numpy()


def images_to_video(frame_dir, video_path, frame_format='frame_%06d.png', input_fps=60, output_fps=60,
                    start_frame=0):
    """Convert the rendered images into a video. The video path format determines whether this will be rendered as
    a GIF or an MP4 (default)."""
    if not os.path.exists(frame_dir):
        raise ValueError("Could not find directory containing frames {}".format(frame_dir))

    is_mp4 = video_path.endswith('.mp4')
    is_gif = video_path.endswith('.gif')
    if not (is_mp4 or is_gif):
        video_path += '.mp4'
        is_mp4 = True

    suffix = '.gif' if is_gif else '.mp4'

    dir_of_file = os.path.dirname(os.path.abspath(video_path))
    if not os.path.exists(dir_of_file):
        os.makedirs(dir_of_file)

    # Make sure we don't override an existing video.
    counter = 0
    video_path_candidate = video_path.replace(suffix, '_{}{}'.format(counter, suffix))
    while os.path.exists(video_path_candidate):
        counter += 1
        video_path_candidate = video_path.replace(suffix, '_{}{}'.format(counter, suffix))

    print("Rendering to video {}".format(os.path.abspath(video_path_candidate)))

    if is_mp4:
        command = ['ffmpeg',
                   '-framerate', str(input_fps),  # must be this early in the command, otherwise it is not applied
                   '-start_number', str(start_frame),
                   '-i', os.path.join(frame_dir, frame_format),
                   '-c:v', 'libx264',
                   '-preset', 'slow',
                   '-profile:v', 'high',
                   '-level:v', '4.0',
                   '-pix_fmt', 'yuv420p',
                   '-vf', 'pad=ceil(iw/2)*2:ceil(ih/2)*2',  # Avoid error when image res is not divisible by 2.
                   '-r', str(output_fps),
                   '-y',
                   video_path_candidate]
    elif is_gif:
        command = ['ffmpeg',
                   '-start_number', str(start_frame),
                   '-framerate', str(input_fps),  # must be this early in the command, otherwise it is not applied
                   '-loglevel', 'panic',
                   '-i', os.path.join(frame_dir, frame_format),
                   '-y',
                   '-filter_complex', "[0:v] fps=15,scale=w=1080:h=-1,split [a][b];[a] palettegen=stats_mode=single [p];[b][p] paletteuse=new=1",
                   '-r', str(output_fps),
                   video_path_candidate]
    else:
        print("Unknown video format, only '.mp4' or '.gif' is supported.")
        return

    FNULL = open(os.devnull, 'w')
    subprocess.Popen(command, stdout=FNULL).wait()
    FNULL.close()


def interpolate_positions(positions, ts_in, ts_out):
    """
    Interpolate positions given at timestamps `ts_in` to timestamps given at `ts_out` with a cubic spline.
    :param positions: A numpy array of shape (F, N, 3), i.e. in angle-axis form.
    :param ts_in: Timestamps corresponding to the given positions, len(ts_in) == F
    :param ts_out: The desired output timestamps.
    :return: A numpy array of shape (len(ts_out), N, 3).
    """
    cs = CubicSpline(ts_in, positions, axis=0)
    new_positions = cs(ts_out)
    return new_positions


def resample_positions(positions, fps_in, fps_out):
    """
    Resample 3D positions from `fps_in` to `fps_out`.
    :param positions: A numpy array of shape (F, ...).
    :param fps_in: The frequency of the input sequence.
    :param fps_out: The desired output frequency.
    :return: A numpy array of shape (F', ...) where F is adjusted according to the new fps.
    """
    n_frames = positions.shape[0]
    assert n_frames > 1, "Resampling with one data point does not make sense."
    duration = n_frames / fps_in
    ts_in = np.arange(0, duration, 1 / fps_in)[:n_frames]
    ts_out = np.arange(0, duration, 1 / fps_out)
    return interpolate_positions(positions, ts_in, ts_out)


def compute_vertex_and_face_normals_torch(vertices, faces, vertex_faces, normalize=False):
    """
    Compute (unnormalized) vertex normals for the given vertices.
    :param vertices: A tensor of shape (N, V, 3).
    :param faces: A tensor of shape (F, 3) indexing into `vertices`.
    :param vertex_faces: A tensor of shape (V, MAX_VERTEX_DEGREE) that lists the face IDs each vertex is a part of.
    :return: The vertex and face normals as tensors of shape (N, V, 3) and (N, F, 3) respectively.
    """
    vs = vertices[:, faces.to(dtype=torch.long)]
    face_normals = torch.cross(vs[:, :, 1] - vs[:, :, 0], vs[:, :, 2] - vs[:, :, 0], dim=-1)  # (N, F, 3)

    ns_all_faces = face_normals[:, vertex_faces]  # (N, V, MAX_VERTEX_DEGREE, 3)
    ns_all_faces[:, vertex_faces == -1] = 0.0
    vertex_degrees = (vertex_faces > -1).sum(dim=-1).to(dtype=ns_all_faces.dtype)
    vertex_normals = ns_all_faces.sum(dim=-2) / vertex_degrees[None, :, None]  # (N, V, 3)

    if normalize:
        face_normals = face_normals / torch.norm(face_normals, dim=-1).unsqueeze(-1)
        vertex_normals = vertex_normals / torch.norm(vertex_normals, dim=-1).unsqueeze(-1)

    return vertex_normals, face_normals


def compute_vertex_and_face_normals(vertices, faces, vertex_faces, normalize=False):
    """
    Compute (unnormalized) vertex normals for the given vertices. This is a rather expensive operation despite it being
    fully optimized with numpy. For a typical SMPL mesh this can take 5-10ms per call. Not normalizing the resulting
    normals speeds things up considerably, so we are omitting this as OpenGL shaders can do this much more efficiently.
    For more speedup, consider caching the result of this function.

    :param vertices: A numpy array of shape (N, V, 3).
    :param faces: A numpy array of shape (F, 3) indexing into `vertices`.
    :param vertex_faces: A numpy array of shape (V, MAX_VERTEX_DEGREE) that lists the face IDs each vertex is a part of.
    :param normalize: Whether to normalize the normals or not.
    :return: The vertex and face normals as a np arrays of shape (N, V, 3) and (N, F, 3) respectively.
    """
    vs = vertices[:, faces]
    face_normals = np.cross(vs[:, :, 1] - vs[:, :, 0], vs[:, :, 2] - vs[:, :, 0], axis=-1)  # (N, F, 3)

    ns_all_faces = face_normals[:, vertex_faces]  # (N, V, MAX_VERTEX_DEGREE, 3)
    ns_all_faces[:, vertex_faces == -1] = 0.0
    vertex_degrees = np.sum(vertex_faces > -1, axis=-1)
    vertex_normals = np.sum(ns_all_faces, axis=-2) / vertex_degrees[np.newaxis, :, np.newaxis]  # (N, V, 3)

    if normalize:
        face_normals = face_normals / np.linalg.norm(face_normals, axis=-1)[..., np.newaxis]
        vertex_normals = vertex_normals / np.linalg.norm(vertex_normals, axis=-1)[..., np.newaxis]

    return vertex_normals, face_normals


def set_lights_in_program(prog, lights):
    """Set program lighting from scene lights"""
    for i, light in enumerate(lights):
        prog['dirLight[{}].pos'.format(i)].value = light.position
        prog['dirLight[{}].color'.format(i)].value = light.color[:3]
        prog['dirLight[{}].intensity_ambient'.format(i)].value = light.intensity_ambient
        prog['dirLight[{}].intensity_diffuse'.format(i)].value = light.intensity_diffuse


def set_material_properties(prog, material):
    prog['diffuse_coeff'].value = material.diffuse
    prog['ambient_coeff'].value = material.ambient


def local_to_global(poses, parents, output_format='aa', input_format='aa'):
    """
    Convert relative joint angles to global ones by unrolling the kinematic chain.
    :param poses: A tensor of shape (N, N_JOINTS*3) defining the relative poses in angle-axis format.
    :param parents: A list of parents for each joint j, i.e. parent[j] is the parent of joint j.
    :param output_format: 'aa' or 'rotmat'.
    :param input_format: 'aa' or 'rotmat'
    :return: The global joint angles as a tensor of shape (N, N_JOINTS*DOF).
    """
    assert output_format in ['aa', 'rotmat']
    assert input_format in ['aa', 'rotmat']
    dof = 3 if input_format == 'aa' else 9
    n_joints = poses.shape[-1] // dof
    if input_format == 'aa':
        local_oris = aa2rot(poses.reshape((-1, 3)))
    else:
        local_oris = poses
    local_oris = local_oris.reshape((-1, n_joints, 3, 3))
    global_oris = torch.zeros_like(local_oris)

    for j in range(n_joints):
        if parents[j] < 0:
            # root rotation
            global_oris[..., j, :, :] = local_oris[..., j, :, :]
        else:
            parent_rot = global_oris[..., parents[j], :, :]
            local_rot = local_oris[..., j, :, :]
            global_oris[..., j, :, :] = torch.matmul(parent_rot, local_rot)

    if output_format == 'aa':
        global_oris = rot2aa(global_oris.reshape((-1, 3, 3)))
        res = global_oris.reshape((-1, n_joints * 3))
    else:
        res = global_oris.reshape((-1, n_joints * 3 * 3))
    return res



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np


def look_at(position, target, up):
    """
    Create an affine transformation that locates the camera at `position`, s.t. it looks at `target`.
    :param position: The 3D position of the camera in world coordinates.
    :param target: The 3D target where the camera should look at in world coordinates.
    :param up: The vector that is considered to be up in world coordinates.
    :return: Returns the 4-by-4 affine transform that transforms a point in world space into the camera space, i.e.
      it returns the inverse of the camera's 6D pose matrix. Assumes right-multiplication, i.e. x' = [R|t] * x.
    """
    def _normalize(x):
        return x / np.linalg.norm(x)

    forward = _normalize(position - target)  # forward actually points in the other direction than `target` is.
    right = _normalize(np.cross(up, forward))
    camera_up = np.cross(forward, right)

    # We directly create the inverse matrix (i.e. world2cam) because this is typically how look-at is define.
    rot = np.eye(4)
    rot[0, :3] = right
    rot[1, :3] = camera_up
    rot[2, :3] = forward

    trans = np.eye(4)
    trans[:3, 3] = -position

    return rot @ trans


def orthographic_projection(scale_x, scale_y, znear, zfar):
    """Returns an orthographic projection matrix."""
    P = np.zeros((4, 4))
    P[0][0] = 1.0 / scale_x
    P[1][1] = 1.0 / scale_y
    P[2][2] = 2.0 / (znear - zfar)
    P[2][3] = (zfar + znear) / (znear - zfar)
    P[3][3] = 1.0
    return P


def perspective_projection(fov, aspect_ratio, znear, zfar):
    """Returns a perspective projection matrix."""
    ar = aspect_ratio
    t = np.tan(fov / 2.0)

    P = np.zeros((4, 4))
    P[0][0] = 1.0 / (ar * t)
    P[1][1] = 1.0 / t
    P[3][2] = -1.0

    f, n = zfar, znear
    if f is None:
        P[2][2] = -1.0
        P[2][3] = -2.0 * n
    else:
        P[2][2] = (f + n) / (n - f)
        P[2][3] = (2 * f * n) / (n - f)

    return P



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.scene.node import Node
from aitviewer.renderables.lines import Lines


class Arrows(Node):
    """
    Draw an arrow.
    """

    def __init__(self,
                 origins,
                 tips,
                 r_base=0.01,
                 r_head=0.02,
                 p=0.1,
                 color=(0.0, 0.0, 1.0, 1.0),
                 **kwargs):
        """
        Initializer.
        :param origins: Set of 3D coordinates of the base of the arrows as a np array of shape (N, B, 3).
        :param tips: Set of 3D coordinates denoting the tip of the arrow (N, T, 3).
        :param r_base: Radius of the base cylinder.
        :param r_head: Radius of the tip cylinder.
        :param p: Percentage of arrow head on the entire length.
        :param color: Color of the line (4-tuple).
        :param mode: 'lines' or 'line_strip' -> ModernGL drawing mode - LINE_STRIP oder LINES
        """
        assert (origins.shape == tips.shape)
        if len(origins.shape) == 2:
            origins = origins[np.newaxis]
            tips = tips[np.newaxis]
        else:
            assert len(origins.shape) == 3
        super(Arrows, self).__init__(n_frames=len(origins), **kwargs)

        self.origins = origins
        self.tips = tips

        # Percentage of arrow head on entire length
        self.p = p

        # Nodes
        self.bases_r = Lines(lines=self.get_line_coords(self.origins, self.mid_points), mode='lines', r_base=r_base,
                             color=color)
        self.arrows_r = Lines(lines=self.get_line_coords(self.mid_points, self.tips), mode='lines', r_base=r_head,
                              r_tip=0.0, color=color)
        self._add_nodes(self.bases_r, self.arrows_r, show_in_hierarchy=False)

    @property
    def mid_points(self):
        return self.origins + (self.tips - self.origins) * (1 - self.p)

    def get_line_coords(self, starts, ends):
        c = np.zeros((len(self), (starts.shape[1] + ends.shape[1]), 3), dtype=starts.dtype)
        c[:, 0::2] = starts
        c[:, 1::2] = ends
        return c

    def redraw(self, **kwargs):
        self.bases_r.lines = self.get_line_coords(self.origins, self.mid_points)
        self.bases_r.redraw(**kwargs)
        self.arrows_r.lines = self.get_line_coords(self.mid_points, self.tips)
        self.arrows_r.redraw(**kwargs)



from setuptools import setup, find_packages

setup(name='aitviewer',
      description='Viewing and rendering of sequences of 3D data.',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url="https://github.com/eth-ait/aitviewer",
      version='1.0.1',
      author='Manuel Kaufmann, Velko Vechev',
      packages=find_packages(),
      include_package_data=True,
      keywords=['viewer', 'moderngl', 'machine learning', 'sequences', 'smpl', 'computer graphics', 'computer vision',
                '3D', 'meshes', 'visualization'],
      platforms=['any'],
      python_requires='>=3.7',
      install_requires=[
              'torch>=1.6.0',
              'numpy>=1.18,<2',
              'opencv-contrib-python-headless>=4.5.1.48',
              'smplx',
              'moderngl-window>=2.4.0',
              'moderngl>=5.6.4,<6.0',
              'PyQt5>=5.15.4',
              'imgui>=1.3.0',
              'tqdm>=4.60.0',
              'trimesh>=3.9.15,<4',
              'rtree>=0.9.7',
              'scipy>=1.5.2',
              'omegaconf>=2.1.1',
              'roma>=1.2.3',
              'joblib'
          ])



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os

import moderngl
import numpy as np
import trimesh
import tqdm

from functools import lru_cache
from PIL import Image

from aitviewer.scene.node import Node
from aitviewer.shaders import get_smooth_lit_with_edges_program
from aitviewer.shaders import get_flat_lit_with_edges_program
from aitviewer.shaders import get_smooth_lit_texturized_program
from aitviewer.utils.utils import compute_vertex_and_face_normals
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties
from trimesh.triangles import points_to_barycentric


class Meshes(Node):
    """A sequence of triangle meshes. This assumes that the mesh topology is fixed over the sequence."""

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 path_to_texture=None,
                 texture_alpha=1.0,
                 **kwargs):
        """
        Initializer.
        :param vertices: A np array of shape (N, V, 3) or (V, 3).
        :param faces: A np array of shape (F, 3).
        :param vertex_normals: A np array of shape (N, V, 3). If not provided, the vertex normals will be computed,
          which incurs some overhead.
        :param face_normals: A np array of shape (N, F, 3). If not provided, the face normals will be computed, which
          incurs some overhead.
        :param vertex_colors: A np array of shape (N, V, 4) overriding the uniform color.
        :param face_colors: A np array of shape (N, F, 4) overriding the uniform or vertex colors.
        :param uv_coords: A np array of shape (V, 2) if the mesh is to be textured.
        :param path_to_texture: Path to an image file that serves as the texture.
        :param texture_alpha: Set transparency for texture.
        """
        if len(vertices.shape) == 2 and vertices.shape[-1] == 3:
            vertices = vertices[np.newaxis]
        assert len(vertices.shape) == 3
        assert len(faces.shape) == 2
        super(Meshes, self).__init__(n_frames=vertices.shape[0], **kwargs)

        self._vertices = vertices
        self.faces = faces.astype(np.int32)

        def _maybe_unsqueeze(x):
            return x[np.newaxis] if x is not None and x.ndim == 2 else x

        self._vertex_normals = _maybe_unsqueeze(vertex_normals)
        self._face_normals = _maybe_unsqueeze(face_normals)
        self.vertex_colors = _maybe_unsqueeze(vertex_colors)
        self.face_colors = _maybe_unsqueeze(face_colors)

        # Texture handling.
        self.has_texture = uv_coords is not None
        self.uv_coords = uv_coords
        self.texture_image = Image.open(path_to_texture).transpose(method=Image.FLIP_TOP_BOTTOM).convert(
            "RGB") if self.has_texture else None
        self.texture_alpha = texture_alpha

        # Misc.
        self.flat_shading = False
        self.show_texture = self.has_texture
        self.norm_coloring = False
        self.normals_r = None


    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        self._vertices = vertices
        self.n_frames = len(vertices)
        # If vertex normals were supplied, they are no longer valid.
        self._vertex_normals = None
        # Must clear all LRU caches where the vertices are used.
        self.compute_vertex_and_face_normals.cache_clear()
        self.redraw()

    @property
    def n_faces(self):
        return self.faces.shape[0]

    @property
    def n_vertices(self):
        return self.vertices.shape[1]

    @property
    def vertex_faces(self):
        # To compute the normals we need to know a mapping from vertex ID to all faces that this vertex is part of.
        # Because we are lazy we abuse trimesh to compute this for us. Not all vertices have the maximum degree, so
        # this array is padded with -1 if necessary.
        return trimesh.Trimesh(self.vertices[0], self.faces, process=False).vertex_faces

    @property
    def vertex_normals(self):
        """Get or compute all vertex normals (this might take a while for long sequences)."""
        if self._vertex_normals is None:
            vertex_normals, _ = compute_vertex_and_face_normals(self.vertices, self.faces, self.vertex_faces,
                                                                normalize=True)
            if vertex_normals.ndim < 3:
                vertex_normals = vertex_normals.unsqueeze(0)
            self._vertex_normals = vertex_normals
        return self._vertex_normals

    @property
    def face_normals(self):
        """Get or compute all face normals (this might take a while for long sequences)."""
        if self._face_normals is None:
            _, face_normals = compute_vertex_and_face_normals(self.vertices, self.faces, self.vertex_faces,
                                                              normalize=True)
            if face_normals.ndim < 3:
                face_normals = face_normals.unsqueeze(0)
            self._face_normals = face_normals
        return self._face_normals

    def vertex_normals_at(self, frame_id):
        """Get or compute the vertex normals at the given frame."""
        if self._vertex_normals is None:
            vn, _ = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._vertex_normals.shape) == 3, f"Got shape {self._vertex_normals.shape}"
            vn = self._vertex_normals[frame_id]
        return vn

    def face_normals_at(self, frame_id):
        """Get or compute the face normals at the given frame."""
        if self._face_normals is None:
            _, fn = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._face_normals.shape) == 3, f"Got shape {self._face_normals.shape}"
            fn = self._face_normals[frame_id]
        return fn

    @property
    def vertex_colors(self):
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, vertex_colors):
        # Vertex colors cannot be empty
        if vertex_colors is None:
            self._vertex_colors = np.full((self.n_frames, self.n_vertices, 4), self.color)
        elif isinstance(vertex_colors, tuple) and len(vertex_colors) == 4:
            self._vertex_colors = np.full((self.n_frames, self.n_vertices, 4), vertex_colors)
        else:
            self._vertex_colors = vertex_colors

    @property
    def face_colors(self):
        return self._face_colors

    @face_colors.setter
    def face_colors(self, face_colors):
        self._face_colors = face_colors
        if face_colors is not None:
            self._vertex_colors = np.tile(self.face_colors[:, :, np.newaxis], [1, 1, 3, 1])
        self.redraw()

    @Node.color.setter
    def color(self, color):
        alpha_changed = np.abs((np.array(color) - np.array(self.color)))[-1] > 0
        self.material.color = color
        if self.is_renderable:
            # If alpha changed, don't update all colors
            if alpha_changed:
                self._vertex_colors[..., -1] = color[-1]
            # Otherwise, if no face colors, then override the existing vertex colors
            elif self.face_colors is None:
                self.vertex_colors = color

        self.redraw()

    @property
    def current_vertices(self):
        return self.vertices[self.current_frame_id]

    @property
    def current_vertex_colors(self):
        return self.vertex_colors[self.current_frame_id]

    def closest_vertex_in_triangle(self, tri_id, point):
        face_vertex_id = np.linalg.norm((self.current_vertices[self.faces[tri_id]] - point), axis=-1).argmin()
        return self.faces[tri_id][face_vertex_id]

    def get_bc_coords_from_points(self, tri_id, points):
        return points_to_barycentric(self.current_vertices[self.faces[[tri_id]]], points)[0]

    @lru_cache(2048)
    def compute_vertex_and_face_normals(self, frame_id, normalize=False):
        """
        Compute face and vertex normals for the given frame. We use an LRU cache since this is a potentially
        expensive operation. This function exists because computing the normals on all frames can increase the
        startup time of the viewer considerably.

        :param frame_id: On which frame to compute the normals.
        :param normalize: Whether or not to normalize the normals. Not doing it is faster and the shaders typically
          enforce unit length of normals anyway.
        :return: The vertex and face normals as a np arrays of shape (V, 3) and (F, 3) respectively.
        """
        vs = self.vertices[frame_id:frame_id + 1]
        vn, fn = compute_vertex_and_face_normals(vs, self.faces, self.vertex_faces, normalize)
        return vn.squeeze(0), fn.squeeze(0)

    @property
    def bounds(self):
        return self.get_bounds(self.vertices)

    def on_frame_update(self):
        """Called whenever a new frame must be displayed."""
        super().on_frame_update()
        self.redraw()

    def redraw(self):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable:
            return

        # Each write call takes about 1-2 ms
        vertices = self.current_vertices
        vertex_colors = self.current_vertex_colors

        if not self.flat_shading:
            vertex_normals = self.vertex_normals_at(self.current_frame_id)
            if self.face_colors is not None:
                vn_for_drawing = vertex_normals[self.faces]
            else:
                vn_for_drawing = vertex_normals
            self.vbo_normals.write(vn_for_drawing.astype('f4').tobytes())

        if self.face_colors is not None:
            vertices = vertices[self.faces]

        self.vbo_vertices.write(vertices.astype('f4').tobytes())
        self.vbo_colors.write(vertex_colors.astype('f4').tobytes())

        if self.has_texture:
            self.texture_prog['texture_alpha'].value = self.texture_alpha
            self.vbo_uvs.write(self.uv_coords.astype('f4').tobytes())

    def _clear_buffer(self):
        self.vbo_vertices.clear()
        self.vbo_colors.clear()
        if self.has_texture:
            self.vbo_uvs.clear()

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        """Prepares this object for rendering. This function must be called before `render` is used."""
        self.smooth_prog = get_smooth_lit_with_edges_program()
        self.flat_prog = get_flat_lit_with_edges_program()

        vertices = self.current_vertices
        vertex_normals = self.vertex_normals_at(self.current_frame_id)
        vertex_colors = self.current_vertex_colors

        # Face colors draw 3 distinct vertices per face.
        if self.face_colors is not None:
            vertices = vertices[self.faces]
            vn_for_drawing = vertex_normals[self.faces]
        else:
            vn_for_drawing = vertex_normals

        self.vbo_vertices = ctx.buffer(vertices.astype('f4').tobytes())
        self.vbo_normals = ctx.buffer(vn_for_drawing.astype('f4').tobytes())
        self.vbo_indices = ctx.buffer(self.faces.tobytes()) if self.face_colors is None else None
        self.vbo_colors = ctx.buffer(vertex_colors.astype('f4').tobytes())

        self.smooth_vao = ctx.vertex_array(self.smooth_prog,
                                           [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                            (self.vbo_normals, '3f4 /v', 'in_normal'),
                                            (self.vbo_colors, '4f4 /v', 'in_color')],
                                           self.vbo_indices)

        self.flat_vao = ctx.vertex_array(self.flat_prog,
                                         [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                          (self.vbo_colors, '4f4 /v', 'in_color')],
                                         self.vbo_indices)

        self.cast_shadow(self.vbo_vertices, self.vbo_indices)

        self.make_pickable(self.vbo_vertices, self.vbo_indices)

        if self.has_texture:
            img = self.texture_image
            self.texture = ctx.texture(img.size, 3, img.tobytes())
            self.texture_prog = get_smooth_lit_texturized_program()
            self.texture_prog['texture_alpha'].value = self.texture_alpha
            self.vbo_uvs = ctx.buffer(self.uv_coords.astype('f4').tobytes())
            self.texture_vao = ctx.vertex_array(self.texture_prog,
                                                [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                                 (self.vbo_normals, '3f4 /v', 'in_normal'),
                                                 (self.vbo_uvs, '2f4 /v', 'in_uv')],
                                                self.vbo_indices)

    def release(self):
        self.smooth_vao.release()
        self.flat_vao.release()
        if self.has_texture:
            self.texture_vao.release()

    def render(self, camera, **kwargs):
        # Check if flat shading changed, in which case we need to update the VBOs.
        flat = kwargs.get('flat_rendering', False)
        if flat != self.flat_shading:
            self.flat_shading = flat
            self.redraw()
        vao = self._prepare_vao(camera, **kwargs)
        vao.render(moderngl.TRIANGLES)

    def _prepare_vao(self, camera, **kwargs):
        """Prepare the shader pipeline and the VAO."""
        if self.has_texture and self.show_texture:
            prog, vao = self.texture_prog, self.texture_vao
            self.texture.use(1)
            prog['diffuse_texture'].value = 1
        else:
            prog, vao = (self.flat_prog, self.flat_vao) if self.flat_shading else (self.smooth_prog, self.smooth_vao)
            prog['norm_coloring'].value = self.norm_coloring

        prog['draw_edges'].value = 1.0 if kwargs['draw_edges'] and self.material._show_edges else 0.0
        prog['win_size'].value = kwargs['window_size']

        self.set_camera_matrices(prog, camera, **kwargs)
        set_lights_in_program(prog, kwargs['lights'])
        set_material_properties(prog, self.material)
        self.receive_shadow(prog, **kwargs)
        return vao

    def _show_normals(self):
        """Create and add normals at runtime"""
        vn = self.vertex_normals
        length = 0.01  # TODO make this either a parameter in the GUI or dependent on the bounding box.
        vn = vn / np.linalg.norm(vn, axis=-1, keepdims=True) * length

        # Must import here because if we do it at the top we create a circular dependency.
        from aitviewer.renderables.arrows import Arrows
        self.normals_r = Arrows(self.vertices, self.vertices + vn,
                                r_base=length / 10, r_head=2 * length / 10, p=0.25, name='Normals')
        self.normals_r.position = self.position
        self.normals_r.current_frame_id = self.current_frame_id
        self.add(self.normals_r, gui_elements=['material'])

    def gui(self, imgui):
        super(Meshes, self).gui(imgui)
        if self.normals_r is None:
            if imgui.button("Show Normals ##s_normals{}".format(self.unique_name)):
                self._show_normals()

            _, self.show_texture = imgui.checkbox('Render Texture', self.show_texture)
            _, self.norm_coloring = imgui.checkbox('Norm Coloring', self.norm_coloring)

            # TODO: Add  export workflow for all nodes
            if imgui.button("Export OBJ"):
                mesh = trimesh.Trimesh(vertices=self.current_vertices, faces=self.faces, process=False)
                mesh.export('../export/' + self.name + '.obj')


class VariableTopologyMeshes(Node):
    """
    A sequence of meshes that do not share the same topology (i.e. different nr of vertices, faces, texture, etc.).
    This simply treats every time instance as a separate mesh, so it's not optimized for performance.
    """

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 texture_paths=None,
                 preload=True,
                 **kwargs):
        """
        Initializer.
        :param vertices: A list of length N with np arrays of size (V_n, 3).
        :param faces: A list of length N with np arrays of shape (F_n, 3).
        :param vertex_normals: An optional list of length N with np arrays of shape (V_n, 3).
        :param face_normals: An optional list of length N with np arrays of shape (F_n, 3).
        :param vertex_colors: An optional list of length N with np arrays of shape (V_n, 4) overriding the uniform color.
        :param face_colors: An optional list of length N with np arrays of shape (F_n, 4) overriding the uniform or vertex colors.
        :param uv_coords: An optional list of length N with np arrays of shape (V_n, 2) if the mesh is to be textured.
        :param texture_paths: An optional list of length N containing paths to the texture as an image file.
        :param preload: Whether or not to pre-load all the meshes. This increases loading time and memory consumption,
          but allows interactive animations.
        """
        assert len(vertices) == len(faces)
        super(VariableTopologyMeshes, self).__init__(n_frames=len(vertices))
        self.preload = preload

        self.vertices = vertices
        self.faces = faces
        self.vertex_normals = vertex_normals
        self.face_normals = face_normals
        self.vertex_colors = vertex_colors
        self.face_colors = face_colors
        self.uv_coords = uv_coords
        self.texture_paths = texture_paths
        self.mesh_kwargs = kwargs

        self._current_mesh = dict()  # maps from frame ID to mesh
        self._all_meshes = []
        if self.preload:
            for f in range(self.n_frames):
                m = self._construct_mesh_at_frame(f)
                self._all_meshes.append(m)

        self.show_texture = True
        self.norm_coloring = False
        self.ctx = None

    def _construct_mesh_at_frame(self, frame_id):
        m = Meshes(self.vertices[frame_id],
                   self.faces[frame_id],
                   self.vertex_normals[frame_id] if self.vertex_normals is not None else None,
                   self.face_normals[frame_id] if self.face_normals is not None else None,
                   self.vertex_colors[frame_id] if self.vertex_colors is not None else None,
                   self.face_colors[frame_id] if self.face_colors is not None else None,
                   self.uv_coords[frame_id] if self.uv_coords is not None else None,
                   self.texture_paths[frame_id] if self.texture_paths is not None else None,
                   **self.mesh_kwargs)
        self.material = m.material
        return m

    @classmethod
    def from_trimeshes(cls, trimeshes, **kwargs):
        """Initialize from a list of trimeshes."""
        vertices = []
        faces = []
        vertex_normals = []
        for m in trimeshes:
            vertices.append(m.vertices)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
        meshes = cls(vertices, faces, vertex_normals, **kwargs)
        return meshes

    @classmethod
    def from_plys(cls, plys, **kwargs):
        """Initialize from a list paths to .ply files."""
        vertices, faces, vertex_normals, vertex_colors = [], [], [], []
        sc = kwargs.get('vertex_scale', 1.0)
        for i in tqdm.tqdm(range(len(plys))):
            m = trimesh.load(plys[i])
            vertices.append(m.vertices * sc)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
            vertex_colors.append(m.visual.vertex_colors / 255.0)
        return cls(vertices, faces, vertex_normals, vertex_colors=vertex_colors)

    @property
    def current_mesh(self):
        if self.preload:
            return self._all_meshes[self.current_frame_id]
        else:
            m = self._current_mesh.get(self.current_frame_id, None)
            if m is None:
                # Need to construct a new one and clean up the old one.
                for k in self._current_mesh:
                    self._current_mesh[k].release()
                m = self._construct_mesh_at_frame(self.current_frame_id)
                m.make_renderable(self.ctx)
                self._current_mesh = {self.current_frame_id: m}
            return m

    @property
    def bounds(self):
        return self.current_mesh.get_bounds(self.current_mesh.vertices)

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        if self.preload:
            for m in self._all_meshes:
                m.make_renderable(ctx)
        else:
            # A bit hacky so that we can make dynamically loaded meshes renderable when we load them.
            self.ctx = ctx
            self.current_mesh.make_renderable(ctx)

    def render(self, camera, **kwargs):
        self.current_mesh.show_texture = self.show_texture
        self.current_mesh.norm_coloring = self.norm_coloring
        self.current_mesh.render(camera, **kwargs)

    def render_shadowmap(self, light_mvp, program):
        self.current_mesh.render_shadowmap(light_mvp, program)

    def gui_position(self, imgui):
        # Position controls
        u, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.current_mesh.position, 0.1, format='%.2f')
        if u:
            for m in self._all_meshes:
                m.position = pos

    def gui_scale(self, imgui):
        # Scale controls
        u, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.current_mesh.scale, 0.01, min_value=0.001,
                                    max_value=10.0, format='%.3f')
        if u:
            self.current_mesh.scale = scale

    def gui_material(self, imgui, show_advanced=True):
        # Color Control
        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.current_mesh.material.color, show_alpha=True)
        if uc:
            for m in self._all_meshes:
                m.color = color

        _, self.show_texture = imgui.checkbox('Render Texture', self.show_texture)
        _, self.norm_coloring = imgui.checkbox('Norm Coloring', self.norm_coloring)

        if show_advanced:
            if imgui.tree_node("Advanced material##advanced_material{}'".format(self.unique_name)):
                # Diffuse
                ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name), self.current_mesh.material.diffuse,
                                                 0.0, 1.0, '%.2f')
                if ud:
                    for m in self._all_meshes:
                        m.material.diffuse = diffuse

                # Ambient
                ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name), self.current_mesh.material.ambient,
                                                 0.0, 1.0, '%.2f')
                if ua:
                    for m in self._all_meshes:
                        m.material.ambient = ambient

                imgui.tree_pop()
